import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt

from pypop7.benchmarks.base_functions import sphere, rosenbrock
from pypop7.optimizers.rs.bes import BES


if __name__ == '__main__':
    sns.set_theme(style='darkgrid')

    problem = {'fitness_function': sphere,
               'ndim_problem': 100}
    options = {'max_function_evaluations': 100*101,
               'seed_rng': 0,  # undefined in the original paper
               'x': np.ones((100,)),
               'n_individuals': 10,
               'saving_fitness': 101}
    bes = BES(problem, options)
    results = bes.optimize()
    print(results)
    plt.figure()
    plt.xlabel('Rounds')
    plt.xticks([0, 20, 40, 60, 80, 100])
    plt.ylabel('Oracle')
    plt.yticks([0, 20, 40, 60, 80, 100])
    plt.plot(results['fitness'][:, 0]/101, results['fitness'][:, 1], 'k')
    plt.show()

    problem = {'fitness_function': rosenbrock,
               'ndim_problem': 100,
               'lower_boundary': -2 * np.ones((100,)),
               'upper_boundary': 2 * np.ones((100,))}
    options = {'max_function_evaluations': 100*101,
               'seed_rng': 0,  # undefined in the original paper
               'n_individuals': 10,
               'c': 0.1,
               'lr': 0.000001,
               'saving_fitness': 101}
    bes = BES(problem, options)
    results = bes.optimize()
    print(results)
    plt.figure()
    plt.xlabel('Rounds')
    plt.xticks([0, 20, 40, 60, 80, 100])
    plt.ylabel('Oracle')
    plt.yticks([10000, 20000, 30000, 40000])
    plt.plot(results['fitness'][:, 0]/101, results['fitness'][:, 1], 'k')
    plt.show()



"""Repeat the following paper for `FEP`:
    Yao, X., Liu, Y. and Lin, G., 1999.
    Evolutionary programming made faster.
    IEEE Transactions on Evolutionary Computation, 3(2), pp.82-102.
    https://ieeexplore.ieee.org/abstract/document/771163

    Since its source code is not openly available till now, the performance differences between our code and
    the original paper are very hard (if not impossible) to analyze. We notice that different people may give
    different implementations based on different understandings of its algorithmic operations, which may
    explain the following performance differences.

    We expect that a much closer open-source implementation could be given in the future, no matter by ourselves
    or others.
"""
import numpy as np

from pypop7.benchmarks.base_functions import sphere, step, rosenbrock, rastrigin, ackley
from pypop7.optimizers.ep.fep import FEP


if __name__ == '__main__':
    ndim_problem = 30

    problem = {'fitness_function': rastrigin,
               'ndim_problem': ndim_problem,
               'lower_boundary': -5.12*np.ones((ndim_problem,)),
               'upper_boundary': 5.12*np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 5000*100,
               'seed_rng': 0,  # undefined in the original paper
               'sigma': 3.0}
    fep = FEP(problem, options)
    results = fep.optimize()
    print(results['best_so_far_y'])
    # 36.358572846518655
    # vs 4.6e-2 (from the original paper)

    problem = {'fitness_function': ackley,
               'ndim_problem': ndim_problem,
               'lower_boundary': -32*np.ones((ndim_problem,)),
               'upper_boundary': 32*np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 1500*100,
               'seed_rng': 0,  # undefined in the original paper
               'sigma': 3.0}
    fep = FEP(problem, options)
    results = fep.optimize()
    print(results['best_so_far_y'])
    # 4.614636065481921
    # vs 1.8e-2 (from the original paper)

    problem = {'fitness_function': sphere,
               'ndim_problem': ndim_problem,
               'lower_boundary': -100*np.ones((ndim_problem,)),
               'upper_boundary': 100*np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 1500*100,
               'seed_rng': 0,  # undefined in the original paper
               'sigma': 3.0}
    fep = FEP(problem, options)
    results = fep.optimize()
    print(results['best_so_far_y'])
    # 0.21750264011392764
    # vs 5.7e-4 (from the original paper)

    problem = {'fitness_function': step,
               'ndim_problem': ndim_problem,
               'lower_boundary': -100*np.ones((ndim_problem,)),
               'upper_boundary': 100*np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 1500*100,
               'seed_rng': 0,  # undefined in the original paper
               'sigma': 3.0}
    fep = FEP(problem, options)
    results = fep.optimize()
    print(results['best_so_far_y'])
    # 38.0
    # vs 0 (from the original paper)

    problem = {'fitness_function': rosenbrock,
               'ndim_problem': ndim_problem,
               'lower_boundary': -30*np.ones((ndim_problem,)),
               'upper_boundary': 30*np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 20000*100,
               'seed_rng': 1,  # undefined in the original paper
               'sigma': 3.0}
    fep = FEP(problem, options)
    results = fep.optimize()
    print(results['best_so_far_y'])
    # 17.962224995187768
    # vs mean 5.06 std 5.87 (from the original paper)



import numpy as np

from pypop7.optimizers.ds.ds import DS


class HJ(DS):
    """Hooke-Jeeves direct (pattern) search method (HJ).

    .. note:: `HJ` is one of the most popular and most cited `DS` methods, originally published in one of
       the top-tier Computer Science journals (i.e., JACM) in 1961. Although sometimes it is still used to optimize
       *low-dimensional* black-box problems, it is **highly recommended** to first attempt other more advanced methods
       for large-scale black-box optimization.

    Parameters
    ----------
    problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`);
              and with the following particular settings (`keys`):
                * 'x'     - initial (starting) point (`array_like`),
                * 'sigma' - initial (global) step-size (`float`),
                * 'gamma' - decreasing factor of step-size (`float`, default: `0.5`).

    Examples
    --------
    Use the Pattern Search optimizer `HJ` to minimize the well-known test function
    `Rosenbrock <http://en.wikipedia.org/wiki/Rosenbrock_function>`_:

    .. code-block:: python
       :linenos:

       >>> import numpy
       >>> from pypop7.benchmarks.base_functions import rosenbrock  # function to be minimized
       >>> from pypop7.optimizers.ds.hj import HJ
       >>> problem = {'fitness_function': rosenbrock,  # define problem arguments
       ...            'ndim_problem': 2,
       ...            'lower_boundary': -5 * numpy.ones((2,)),
       ...            'upper_boundary': 5 * numpy.ones((2,))}
       >>> options = {'max_function_evaluations': 5000,  # set optimizer options
       ...            'seed_rng': 2022,
       ...            'x': 3 * numpy.ones((2,)),
       ...            'sigma': 0.1,
       ...            'verbose_frequency': 500}
       >>> hj = HJ(problem, options)  # initialize the optimizer class
       >>> results = hj.optimize()  # run the optimization process
       >>> # return the number of function evaluations and best-so-far fitness
       >>> print(f"HJ: {results['n_function_evaluations']}, {results['best_so_far_y']}")
       HJ: 5000, 0.22119484961034389

    Furthermore, an interesting visualization of `HJ`'s search trajectory on a 2-dimensional test function is shown in
    `this GitHub link <https://github.com/Evolutionary-Intelligence/pypop/blob/main/docs/demo/demo_hj.gif>`_.

    Attributes
    ----------
    x     : `array_like`
            starting search point.
    sigma : `float`
            final (global) step-size.
    gamma : `float`
            decreasing factor of step-size.

    References
    ----------
    Kochenderfer, M.J. and Wheeler, T.A., 2019.
    Algorithms for optimization.
    MIT Press.
    https://algorithmsbook.com/optimization/files/chapter-7.pdf
    (See Algorithm 7.5 (Page 104) for details.)

    Kaupe Jr, A.F., 1963.
    Algorithm 178: Direct search.
    Communications of the ACM, 6(6), pp.313-314.
    https://dl.acm.org/doi/pdf/10.1145/366604.366632

    Hooke, R. and Jeeves, T.A., 1961.
    “Direct search” solution of numerical and statistical problems.
    Journal of the ACM, 8(2), pp.212-229.
    https://dl.acm.org/doi/10.1145/321062.321069
    """
    def __init__(self, problem, options):
        DS.__init__(self, problem, options)
        self.gamma = options.get('gamma', 0.5)  # decreasing factor of step-size (γ)
        assert self.gamma > 0.0, f'`self.gamma` == {self.gamma}, but should > 0.0.'

    def initialize(self, args=None, is_restart=False):
        x = self._initialize_x(is_restart)  # initial point
        y = self._evaluate_fitness(x, args)  # fitness
        return x, y

    def iterate(self, args=None, x=None, fitness=None):
        improved, best_so_far_x, best_so_far_y = False, self.best_so_far_x, self.best_so_far_y
        for i in range(self.ndim_problem):  # to search along each coordinate
            for sgn in [-1, 1]:  # for two opponent directions
                if self._check_terminations():
                    return None
                xx = np.copy(best_so_far_x)
                xx[i] += sgn*self.sigma
                y = self._evaluate_fitness(xx, args)
                if self.saving_fitness:
                    fitness.append(y)
                if y < best_so_far_y:
                    best_so_far_y, improved = y, True
        if not improved:  # to decrease step-size if no improvement
            self.sigma *= self.gamma  # alpha
        return None

    def restart_reinitialize(self, args=None, x=None, y=None, fitness=None):
        self._fitness_list.append(self.best_so_far_y)
        is_restart_1, is_restart_2 = self.sigma < self.sigma_threshold, False
        if len(self._fitness_list) >= self.stagnation:
            is_restart_2 = (self._fitness_list[-self.stagnation] - self._fitness_list[-1]) < self.fitness_diff
        is_restart = bool(is_restart_1) or bool(is_restart_2)
        if is_restart:
            self.sigma = np.copy(self._sigma_bak)
            x, y = self.initialize(args, is_restart)
            if self.saving_fitness:
                fitness.append(y)
            self._fitness_list = [self.best_so_far_y]
            self._n_generations = 0
            self._n_restart += 1
            if self.verbose:
                print(' ....... restart .......')
            self._print_verbose_info(y)
        return x, y

    def optimize(self, fitness_function=None, args=None):
        fitness = DS.optimize(self, fitness_function)
        x, y = self.initialize(args)
        if self.saving_fitness:
            fitness.append(y)
        self._print_verbose_info(y)
        while True:
            self.iterate(args, x, fitness)
            if self._check_terminations():
                break
            self._n_generations += 1
            self._print_verbose_info(y)
            if self.is_restart:
                x, y = self.restart_reinitialize(args, x, y, fitness)
        return self._collect_results(fitness)



import numpy as np

from pypop7.optimizers.core.optimizer import Optimizer


class EP(Optimizer):
    """Evolutionary Programming (EP).

    This is the **base** (abstract) class for all `EP` classes. Please use any of its instantiated subclasses
    to optimize the black-box problem at hand.

    .. note:: `EP` is one of three classical families of evolutionary algorithms (EAs), proposed originally by Lawrence
    J. Fogel, recipient of both `Evolutionary Computation Pioneer Award 1998 <https://tinyurl.com/456as566>`_ and
    `IEEE Frank Rosenblatt Award 2006 <https://tinyurl.com/yj28zxfa>`_. When used for continuous optimization, most
    of modern `EP` share much similarities (e.g. self-adaptation) with `ES`, another representative EA.

    Parameters
    ----------
    problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`);
              and with the following particular settings (`keys`):
                * 'sigma'         - initial global step-size, mutation strength (`float`),
                * 'n_individuals' - number of offspring, offspring population size (`int`, default: `100`).

    Attributes
    ----------
    n_individuals : `int`
                    number of offspring, offspring population size.
    sigma         : `float`
                    initial global step-size, mutation strength.

    Methods
    -------

    References
    ----------
    Yao, X., Liu, Y. and Lin, G., 1999.
    Evolutionary programming made faster.
    IEEE Transactions on Evolutionary Computation, 3(2), pp.82-102.
    https://ieeexplore.ieee.org/abstract/document/771163
    """
    def __init__(self, problem, options):
        Optimizer.__init__(self, problem, options)
        if self.n_individuals is None:
            self.n_individuals = 100  # number of offspring, offspring population size
        self.sigma = options.get('sigma')  # initial global step-size, mutation strength
        self._n_generations = 0  # number of generations

    def initialize(self):
        raise NotImplementedError

    def iterate(self):
        raise NotImplementedError

    def _print_verbose_info(self, y):
        if self.verbose and (not self._n_generations % self.verbose):
            info = '  * Generation {:d}: best_so_far_y {:7.5e}, min(y) {:7.5e} & Evaluations {:d}'
            print(info.format(self._n_generations, self.best_so_far_y, np.min(y), self.n_function_evaluations))

    def _collect_results(self, fitness=None):
        results = Optimizer._collect_results(self, fitness)
        results['_n_generations'] = self._n_generations
        return results



import numpy as np

from pypop7.optimizers.core.optimizer import Optimizer


class CEM(Optimizer):
    """Cross-Entropy Method (CEM).

    This is the **base** (abstract) class for all `CEM` classes. Please use any of its instantiated subclasses to
    optimize the black-box problem at hand.

    .. note:: `CEM` is a class of well-established, principled population-based optimizers, proposed originally
       by Rubinstein, whose core idea is based on **Kullback–Leibler (or cross-entropy) minimization**.

       "CEM is not only based on fundamental principles (cross-entropy distance, maximum likelihood, etc.), but is
       also very easy to program (with far fewer parameters than many other global optimization heuristics), and
       gives consistently accurate results, and is therefore worth considering when faced with a difficult
       optimization problem."---[Kroese et al., 2006, MCAP]

    References
    ----------
    Rubinstein, R.Y. and Kroese, D.P., 2016.
    Simulation and the Monte Carlo method (Third Edition).
    John Wiley & Sons.
    https://onlinelibrary.wiley.com/doi/book/10.1002/9781118631980

    Kroese, D.P., Porotsky, S. and Rubinstein, R.Y., 2006.
    The cross-entropy method for continuous multi-extremal optimization.
    Methodology and Computing in Applied Probability, 8(3), pp.383-407.
    https://link.springer.com/article/10.1007/s11009-006-9753-0

    De Boer, P.T., Kroese, D.P., Mannor, S. and Rubinstein, R.Y., 2005.
    A tutorial on the cross-entropy method.
    Annals of Operations Research, 134(1), pp.19-67.
    https://link.springer.com/article/10.1007/s10479-005-5724-z

    Rubinstein, R.Y. and Kroese, D.P., 2004.
    The cross-entropy method: a unified approach to combinatorial optimization, Monte-Carlo simulation,
    and machine learning.
    New York: Springer.
    https://link.springer.com/book/10.1007/978-1-4757-4321-0
    """
    def __init__(self, problem, options):
        Optimizer.__init__(self, problem, options)
        if self.n_individuals is None:  # number of individuals (samples)
            self.n_individuals = 1000
        if self.n_parents is None:  # number of elitists
            self.n_parents = 200
        self.mean = options.get('mean')  # mean of Gaussian search (sampling/mutation) distribution
        if self.mean is None:
            self.mean = options.get('x')
        self.sigma = options.get('sigma')  # global (overall) step-size
        assert self.sigma is not None
        self._sigmas = self.sigma*np.ones((self.ndim_problem,))  # individual step-sizes
        self._n_generations = 0

    def initialize(self):
        raise NotImplementedError

    def iterate(self):
        raise NotImplementedError

    def _initialize_mean(self, is_restart=False):
        if is_restart or (self.mean is None):
            mean = self.rng_initialization.uniform(self.initial_lower_boundary, self.initial_upper_boundary)
        else:
            mean = np.copy(self.mean)
        return mean

    def _print_verbose_info(self, y):
        if self.verbose and (not self._n_generations % self.verbose):
            info = '  * Generation {:d}: best_so_far_y {:7.5e}, min(y) {:7.5e} & Evaluations {:d}'
            print(info.format(self._n_generations, self.best_so_far_y, np.min(y), self.n_function_evaluations))

    def _collect_results(self, fitness, mean=None):
        results = Optimizer._collect_results(self, fitness)
        results['mean'] = mean
        results['_sigmas'] = self._sigmas
        results['_n_generations'] = self._n_generations
        return results



import numpy as np

from pypop7.optimizers.pso.pso import PSO


class SPSO(PSO):
    """Standard Particle Swarm Optimizer with a global topology (SPSO).

    .. note:: *"In the case of multidimensional functions and of functions that are defined over discrete or
       non-Euclidean spaces, one must find the most appropriate ways of computing directions and updating
       velocities so that particles converge toward the optimum of the function."*
       ---[Floreano&Mattiussi, 2008]

    Parameters
    ----------
    problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`);
              and with the following particular settings (`keys`):
                * 'n_individuals' - swarm (population) size, number of particles (`int`, default: `20`),
                * 'cognition'     - cognitive learning rate (`float`, default: `2.0`),
                * 'society'       - social learning rate (`float`, default: `2.0`),
                * 'max_ratio_v'   - maximal ratio of velocities w.r.t. search range (`float`, default: `0.2`).

    Examples
    --------
    Use the PSO optimizer `SPSO` to minimize the well-known test function
    `Rosenbrock <http://en.wikipedia.org/wiki/Rosenbrock_function>`_:

    .. code-block:: python
       :linenos:

       >>> import numpy
       >>> from pypop7.benchmarks.base_functions import rosenbrock  # function to be minimized
       >>> from pypop7.optimizers.pso.spso import SPSO
       >>> problem = {'fitness_function': rosenbrock,  # define problem arguments
       ...            'ndim_problem': 2,
       ...            'lower_boundary': -5 * numpy.ones((2,)),
       ...            'upper_boundary': 5 * numpy.ones((2,))}
       >>> options = {'max_function_evaluations': 5000,  # set optimizer options
       ...            'seed_rng': 2022}
       >>> spso = SPSO(problem, options)  # initialize the optimizer class
       >>> results = spso.optimize()  # run the optimization process
       >>> # return the number of function evaluations and best-so-far fitness
       >>> print(f"SPSO: {results['n_function_evaluations']}, {results['best_so_far_y']}")
       SPSO: 5000, 3.456165303371902e-09

    Attributes
    ----------
    n_individuals : `int`
                    swarm (population) size, number of particles.
    cognition     : `float`
                    cognitive learning rate.
    society       : `float`
                    social learning rate.
    max_ratio_v   : `float`
                    maximal ratio of velocities w.r.t. search range.

    References
    ----------
    Floreano, D. and Mattiussi, C., 2008.
    Bio-inspired artificial intelligence: Theories, methods, and technologies.
    MIT Press.
    https://mitpress.mit.edu/9780262062718/bio-inspired-artificial-intelligence/
    (See [Chapter 7.2 Particle Swarm Optimization] for details.)

    Venter, G. and Sobieszczanski-Sobieski, J., 2003.
    Particle swarm optimization.
    AIAA Journal, 41(8), pp.1583-1589.
    https://arc.aiaa.org/doi/abs/10.2514/2.2111

    Eberhart, R.C., Shi, Y. and Kennedy, J., 2001.
    Swarm intelligence.
    Elsevier.
    https://www.elsevier.com/books/swarm-intelligence/eberhart/978-1-55860-595-4

    Shi, Y. and Eberhart, R., 1998, May.
    A modified particle swarm optimizer.
    In IEEE World Congress on Computational Intelligence (pp. 69-73). IEEE.
    https://ieeexplore.ieee.org/abstract/document/699146

    Kennedy, J. and Eberhart, R., 1995, November.
    Particle swarm optimization.
    In Proceedings of International Conference on Neural Networks (pp. 1942-1948). IEEE.
    https://ieeexplore.ieee.org/document/488968
    """
    def __init__(self, problem, options):
        PSO.__init__(self, problem, options)

    def iterate(self, v=None, x=None, y=None, p_x=None, p_y=None, n_x=None, args=None):
        for i in range(self.n_individuals):
            if self._check_terminations():
                return v, x, y, p_x, p_y, n_x
            n_x[i] = p_x[np.argmin(p_y)]  # online update within global topology
            cognition_rand = self.rng_optimization.uniform(size=(self.ndim_problem,))
            society_rand = self.rng_optimization.uniform(size=(self.ndim_problem,))
            v[i] = (self._w[self._n_generations]*v[i] +
                    self.cognition*cognition_rand*(p_x[i] - x[i]) +
                    self.society*society_rand*(n_x[i] - x[i]))  # velocity update
            min_v, max_v = v[i] < self._min_v, v[i] > self._max_v
            v[i, min_v], v[i, max_v] = self._min_v[min_v], self._max_v[max_v]
            x[i] += v[i]  # position update
            y[i] = self._evaluate_fitness(x[i], args)  # fitness evaluation
            if y[i] < p_y[i]:  # online update
                p_x[i], p_y[i] = x[i], y[i]
        self._n_generations += 1
        return v, x, y, p_x, p_y, n_x



import numpy as np

from pypop7.optimizers.es.es import ES


class CCMAES(ES):
    """Cholesky-CMA-ES (CCMAES, (μ/μ_w,λ)-Cholesky-CMA-ES).

    Reference
    ---------
    Suttorp, T., Hansen, N. and Igel, C., 2009.
    Efficient covariance matrix update for variable metric evolution strategies.
    Machine Learning, 75(2), pp.167-197.
    https://link.springer.com/article/10.1007/s10994-009-5102-1
    (See Algorithm 4 for details.)
    """
    def __init__(self, problem, options):
        ES.__init__(self, problem, options)
        self.c_s = options.get('c_s', self._set_c_s())
        self.d_s = options.get('d_s', self._set_d_s())
        self.c_c = options.get('c_c', 4 / (self.ndim_problem + 4))
        self.c_cov = options.get('c_cov', 2 / np.power(self.ndim_problem + np.sqrt(2), 2))

    def _set_c_s(self):
        return np.sqrt(self._mu_eff) / (np.sqrt(self.ndim_problem) + np.sqrt(self._mu_eff))

    def _set_d_s(self):
        return 1 + 2 * np.maximum(0, np.sqrt((self._mu_eff - 1) / (self.ndim_problem + 1)) - 1) + self.c_s

    def initialize(self, is_restart=False):
        mean = self._initialize_mean(is_restart)  # mean of Gaussian search distribution
        z = np.empty((self.n_individuals, self.ndim_problem))  # Gaussian noise for mutation
        x = np.empty((self.n_individuals, self.ndim_problem))  # offspring population
        a = np.diag(np.ones(self.ndim_problem,))  # Cholesky factors
        a_i = np.diag(np.ones(self.ndim_problem,))  # inverse of Cholesky factors
        p_s = np.zeros((self.ndim_problem,))  # evolution path for global step-size adaptation
        p_c = np.zeros((self.ndim_problem,))  # evolution path for covariance matrix adaptation
        y = np.empty((self.n_individuals,))  # fitness (no evaluation)
        if is_restart:
            self.c_s = self._set_c_s()
            self.d_s = self._set_d_s()
        return mean, z, x, a, a_i, p_s, p_c, y

    def iterate(self, z=None, x=None, mean=None, a=None, y=None, args=None):
        for k in range(self.n_individuals):  # Line 4
            if self._check_terminations():
                return z, x, y
            z[k] = self.rng_optimization.standard_normal((self.ndim_problem,))  # Line 5
            x[k] = mean + self.sigma * np.dot(a, z[k])  # Line 6
            y[k] = self._evaluate_fitness(x[k], args)
        return z, x, y

    def _update_distribution(self, z=None, x=None, a=None, a_i=None, p_s=None, p_c=None, y=None):
        order = np.argsort(y)[:self.n_parents]
        mean, z_w = np.dot(self._w, x[order]), np.dot(self._w, z[order])  # Line 3 and 7
        p_c = (1 - self.c_c) * p_c + np.sqrt(self.c_c * (2 - self.c_c) * self._mu_eff) * np.dot(a, z_w)  # Line 8
        v = np.dot(a_i, p_c)  # Line 9
        v_norm = np.dot(v, v)  # (||v||)^2
        s_v_norm = np.sqrt(1 + self.c_cov / (1 - self.c_cov) * v_norm)
        a_i = (a_i - (1 - 1 / s_v_norm) * np.dot(v[:, np.newaxis], np.dot(v[np.newaxis, :], a_i)) / v_norm
               ) / np.sqrt(1 - self.c_cov)  # Line 10
        a = np.sqrt(1 - self.c_cov) * (a + (s_v_norm - 1) * np.dot(
            p_c[:, np.newaxis], v[np.newaxis, :]) / v_norm)  # Line 11
        p_s = (1 - self.c_s) * p_s + np.sqrt(self.c_s * (2 - self.c_s) * self._mu_eff) * z_w  # Line 12
        self.sigma *= np.exp(self.c_s / self.d_s * (np.linalg.norm(p_s) / self._e_chi - 1))  # Line 13
        return mean, a, a_i, p_s, p_c

    def restart_initialize(self, mean=None, z=None, x=None, a=None, a_i=None, p_s=None, p_c=None, y=None):
        is_restart = ES.restart_initialize(self)
        if is_restart:
            mean, z, x, a, a_i, p_s, p_c, y = self.initialize(True)
        return mean, z, x, a, a_i, p_s, p_c, y

    def optimize(self, fitness_function=None, args=None):  # for all generations (iterations)
        fitness = ES.optimize(self, fitness_function)
        mean, z, x, a, a_i, p_s, p_c, y = self.initialize()
        while True:
            # sample and evaluate offspring population
            z, x, y = self.iterate(z, x, mean, a, y, args)
            if self.record_fitness:
                fitness.extend(y)
            if self._check_terminations():
                break
            mean, a, a_i, p_s, p_c = self._update_distribution(z, x, a, a_i, p_s, p_c, y)
            self._n_generations += 1
            self._print_verbose_info(y)
            if self.is_restart:
                mean, z, x, a, a_i, p_s, p_c, y = self.restart_initialize(mean, z, x, a, a_i, p_s, p_c, y)
        results = self._collect_results(fitness, mean)
        return results



import numpy as np

from pypop7.optimizers.es.es import ES


class DSAES(ES):
    """Derandomized Self-Adaptation Evolution Strategy (DSAES).

    .. note:: `DSAES` adapts the *individual* step-sizes on-the-fly with *small* populations. To obtain
       satisfactory performance for large-scale black-box optimization, the number of offspring may
       need to be carefully tuned.

    Parameters
    ----------
    problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`);
              and with the following particular settings (`keys`):
                * 'mean'          - initial (starting) point, mean of Gaussian search distribution (`array_like`),
                * 'sigma'         - initial global step-size (σ), mutation strength (`float`),
                * 'n_individuals' - number of offspring (λ: lambda), offspring population size (`int`, default: `10`),
                * 'lr_sigma'      - learning rate of global step-size (`float`, default: `1.0/3.0`).

    Examples
    --------
    Use the ES optimizer `DSAES` to minimize the well-known test function
    `Rosenbrock <http://en.wikipedia.org/wiki/Rosenbrock_function>`_:

    .. code-block:: python
       :linenos:

       >>> import numpy
       >>> from pypop7.benchmarks.base_functions import rosenbrock  # function to be minimized
       >>> from pypop7.optimizers.es.dsaes import DSAES
       >>> problem = {'fitness_function': rosenbrock,  # define problem arguments
       ...            'ndim_problem': 2,
       ...            'lower_boundary': -5 * numpy.ones((2,)),
       ...            'upper_boundary': 5 * numpy.ones((2,))}
       >>> options = {'max_function_evaluations': 5000,  # set optimizer options
       ...            'seed_rng': 2022,
       ...            'mean': 3 * numpy.ones((2,)),
       ...            'sigma': 0.1}
       >>> dsaes = DSAES(problem, options)  # initialize the optimizer class
       >>> results = dsaes.optimize()  # run the optimization process
       >>> # return the number of function evaluations and best-so-far fitness
       >>> print(f"DSAES: {results['n_function_evaluations']}, {results['best_so_far_y']}")
       DSAES: 5000, 0.04805047881994932

    Attributes
    ----------
    n_individuals : `int`
                    number of offspring (λ: lambda), offspring population size.
    mean          : `array_like`
                    mean of Gaussian search distribution.
    sigma         : `float`
                    mutation strength.
    lr_sigma      : `float`
                    learning rate of global step-size.

    References
    ----------
    Hansen, N., Arnold, D.V. and Auger, A., 2015.
    Evolution strategies.
    In Springer Handbook of Computational Intelligence (pp. 871-898). Springer, Berlin, Heidelberg.
    https://link.springer.com/chapter/10.1007%2F978-3-662-43505-2_44
    (See Algorithm 44.5 for details.)

    Ostermeier, A., Gawelczyk, A. and Hansen, N., 1994.
    A derandomized approach to self-adaptation of evolution strategies.
    Evolutionary Computation, 2(4), pp.369-380.
    https://direct.mit.edu/evco/article-abstract/2/4/369/1407/A-Derandomized-Approach-to-Self-Adaptation-of
    """
    def __init__(self, problem, options):
        if options.get('n_individuals') is None:
            options['n_individuals'] = 10
        ES.__init__(self, problem, options)
        if self.lr_sigma is None:
            self.lr_sigma = 1.0/3.0
        self._axis_sigmas = None
        self._e_hnd = np.sqrt(2.0/np.pi)  # E[|N(0,1)|]: expectation of half-normal distribution

    def initialize(self, is_restart=False):
        self._axis_sigmas = self._sigma_bak*np.ones((self.ndim_problem,))
        x = np.empty((self.n_individuals, self.ndim_problem))  # offspring population
        mean = self._initialize_mean(is_restart)  # mean of Gaussian search distribution
        # set individual step-sizes for all offspring
        sigmas = np.ones((self.n_individuals, self.ndim_problem))
        y = np.empty((self.n_individuals,))  # fitness (no evaluation)
        return x, mean, sigmas, y

    def iterate(self, x=None, mean=None, sigmas=None, y=None, args=None):
        # sample offspring population (Line 4)
        for k in range(self.n_individuals):
            if self._check_terminations():
                return x, sigmas, y
            sigma = self.lr_sigma*self.rng_optimization.standard_normal()  # Line 5
            z = self.rng_optimization.standard_normal((self.ndim_problem,))  # Line 6
            x[k] = mean + np.exp(sigma)*self._axis_sigmas*z  # Line 7
            # mimick the effect of intermediate recombination (Line 8)
            sigmas_1 = np.power(np.exp(np.abs(z)/self._e_hnd - 1.0), 1.0/self.ndim_problem)
            sigmas_2 = np.power(np.exp(sigma), 1.0/np.sqrt(self.ndim_problem))
            sigmas[k] = self._axis_sigmas*sigmas_1*sigmas_2
            y[k] = self._evaluate_fitness(x[k], args)
        return x, sigmas, y

    def _restart_reinitialize(self):
        self._fitness_list.append(self.best_so_far_y)
        is_restart_1, is_restart_2 = np.all(self._axis_sigmas < self.sigma_threshold), False
        if len(self._fitness_list) >= self.stagnation:
            is_restart_2 = (self._fitness_list[-self.stagnation] - self._fitness_list[-1]) < self.fitness_diff
        is_restart = bool(is_restart_1) or bool(is_restart_2)
        if is_restart:
            self._n_restart += 1
            self.n_individuals *= 2
            self._n_generations = 0
            self._fitness_list = [np.Inf]
            if self.verbose:
                print(' ...*... restart ...*...')
        return is_restart

    def restart_reinitialize(self, x=None, mean=None, sigmas=None, y=None):
        if self._restart_reinitialize():
            x, mean, sigmas, y = self.initialize(True)
        return x, mean, sigmas, y

    def optimize(self, fitness_function=None, args=None):  # for all generations (iterations)
        fitness = ES.optimize(self, fitness_function)
        x, mean, sigmas, y = self.initialize()
        while True:
            # sample and evaluate offspring population
            x, sigmas, y = self.iterate(x, mean, sigmas, y, args)
            if self.saving_fitness:
                fitness.extend(y)
            if self._check_terminations():
                break
            order = np.argsort(y)[0]  # Line 9
            self._axis_sigmas = np.copy(sigmas[order])  # Line 10
            mean = np.copy(x[order])  # Line 11
            self._print_verbose_info(y)
            self._n_generations += 1
            if self.is_restart:
                x, mean, sigmas, y = self.restart_reinitialize(x, mean, sigmas, y)
        results = self._collect_results(fitness, mean)
        results['_axis_sigmas'] = self._axis_sigmas
        return results



import numpy as np

from pypop7.optimizers.cem.cem import CEM
import colorednoise


class SECEM(CEM):
    """Sample-efficient Cross-Entropy Method

    Parameters
    ----------
    problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).

    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`),
              and with the following particular settings (`keys`):
                * 'n_individuals' - population size (`int`, default: `1000`),
                * 'n_parents' - parent size (`int`, default: `200`),
                * 'mean' - initial mean value (`array_like`, default: `4 * np.ones((ndim_problem,))`),
                * 'sigma' - initial global step-size (σ), mutation strength (`float`, default: '1.0'),
                * 'fraction_reused_elites' - reuse fraction (`float`, default: `0.3`),
                * 'noise_beta' - noise parameter (`float`, default: `2.0`),

    Examples
    --------
    Use the CEM optimizer `SECEM` to minimize the well-known test function
    `Rosenbrock <http://en.wikipedia.org/wiki/Rosenbrock_function>`_:

    .. code-block:: python
       :linenos:

       >>> import numpy
       >>> from pypop7.benchmarks.base_functions import rosenbrock  # function to be minimized
       >>> from pypop7.optimizers.cem.secem import SECEM
       >>> problem = {'fitness_function': rosenbrock,  # define problem arguments
       ...            'ndim_problem': 100,
       ...            'lower_boundary': -5 * numpy.ones((100,)),
       ...            'upper_boundary': 5 * numpy.ones((100,))}
       >>> options = {'max_function_evaluations': 1000000,  # set optimizer options
       ...            'n_individuals': 100,
       ...            'n_parents': 20,
       ...            'mean': 4 * np.ones((100,)),
       ...            'fraction_reused_elites': 0.3,
       ...            'sigma': 0.5,
       ...            'noise_beta': 2.0,
       ...            'seed_rng': 2022}
       >>> secem = SECEM(problem, options)  # initialize the optimizer class
       >>> results = secem.optimize()  # run the optimization process
       >>> # return the number of function evaluations and best-so-far fitness
       >>> print(f"SECEM: {results['n_function_evaluations']}, {results['best_so_far_y']}")
       SECEM: 1000000, 1.7065180618298407e-05

    Attributes
    ----------
    n_individuals           : `int`
                              number of offspring (λ: lambda), offspring population size.
    n_parents               : `int`
                              number of parents (μ: mu), parental population size.
    mean                    : `array_like`
                              mean of Gaussian search distribution.
    sigma                   : `float`
                              mutation strength.
    fraction_reused_elites  : `float`
                              reuse fraction
    noise_beta              : `float`
                              noise parameter

    Reference
    ---------------
    C. Pinneri, S. Sawant, S. Blaes, J. Achterhold, J. Stuckler, M. Rolinek, G. Martius
    Sample-efficient Cross-Entropy Method for Real-time Planning
    https://arxiv.org/pdf/2008.06389.pdf
    """
    def __init__(self, problem, options):
        CEM.__init__(self, problem, options)
        self.alpha = 0.1
        self.fraction_reused_elites = options.get('fraction_reused_elites', 0.3)
        self.decay = 1.25  # decay parameter used in N
        self.noise_beta = options.get('noise_beta', 2.0)

    def initialize(self, is_restart=False):
        mean = self._initialize_mean(is_restart)
        y = np.empty((self.n_individuals,))
        x = np.empty((self.n_individuals, self.ndim_problem))
        elite_x = np.empty((self.n_parents, self.ndim_problem))
        elite_y = np.empty((self.n_parents,))
        return mean, x, elite_x, y, elite_y

    def re_initialize(self):
        y = np.empty((self.n_individuals,))
        return y

    def iterate(self, mean, x, y):
        for i in range(self.n_individuals):
            x[i] = colorednoise.powerlaw_psd_gaussian(self.noise_beta, size=(1, self.ndim_problem))
            x[i] = np.clip(np.dot(x[i], self.sigma) + self.mean, self.lower_boundary, self.upper_boundary)
            y[i] = self._evaluate_fitness(x[i])
            if self._check_terminations():
                return x, y
        return x, y

    def _update_parameters(self, mean, x, elite_x, y, elite_y):
        if self._n_generations == 0:
            order = np.argsort(y)
            for j in range(self.n_parents):
                elite_x[j] = x[order[j]]
                elite_y[j] = y[order[j]]
        else:
            length = int(self.fraction_reused_elites * self.n_parents)
            combine_x = np.empty((length, self.ndim_problem))
            combine_y = np.empty((length,))
            for k in range(length):
                combine_x[k] = elite_x[k]
                combine_y[k] = elite_y[k]
            combine_x = np.vstack((x, combine_x))
            combine_y = np.hstack((y, combine_y))
            order = np.argsort(combine_y)
            for j in range(self.n_parents):
                elite_x[j] = combine_x[order[j]]
                elite_y[j] = combine_y[order[j]]
        mean = self.alpha * mean + (1 - self.alpha) * np.mean(elite_x, axis=0)
        self.sigma = self.alpha * self.sigma + (1 - self.alpha) * np.std(elite_x, axis=0)
        return mean, elite_x, elite_y

    def optimize(self, fitness_function=None, args=None):
        fitness = CEM.optimize(self, fitness_function)
        mean, x, elite_x, y, elite_y = self.initialize()
        restart = True
        while True:
            self.n_individuals = int(max(self.n_individuals * np.power(self.decay, -1 * self._n_generations),
                                     2 * self.n_parents))
            if self.n_individuals != 2 * self.n_parents or restart is True:
                y = self.re_initialize()
                if self.n_individuals == 2 * self.n_parents:
                    restart = False
            x, y = self.iterate(mean, x, y)
            if self.saving_fitness:
                fitness.extend(y)
            if self._check_terminations():
                break
            mean, elite_x, elite_y = self._update_parameters(mean, x, elite_x, y, elite_y)
            self._n_generations += 1
            self._print_verbose_info(y)
        results = self._collect_results(fitness, mean)
        return results



"""Repeat the following paper for `ASGA`:
    Demo, N., Tezzele, M. and Rozza, G., 2021.
    A supervised learning approach involving active subspaces for an efficient genetic algorithm in high-dimensional
    optimization problems.
    SIAM Journal on Scientific Computing, 43(3), pp.B831-B853.
    https://epubs.siam.org/doi/10.1137/20M1345219

    Very close performance can be obtained by our code. Therefore, we argue that
    the repeatability of `JADE` can be well-documented (*at least partly*).
"""
import numpy as np

from pypop7.benchmarks.base_functions import sphere, step, rosenbrock, rastrigin, ackley
from pypop7.optimizers.ga.asga import ASGA


if __name__ == '__main__':
    ndim_problem = 15

    problem = {'fitness_function': sphere,
               'ndim_problem': ndim_problem,
               'lower_boundary': -100*np.ones((ndim_problem,)),
               'upper_boundary': 100*np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 1500 * 20,
               'n_initial_individuals': 2000,
               'n_individuals': 200,
               'seed_rng': 0,  # undefined in the original paper
               }
    asga = ASGA(problem, options)
    results = asga.optimize()
    print(results)
    print(results['best_so_far_y'])
    # evaluations 3800
    # 3.07029e-20
    # vs (from the original paper)



import numpy as np

from pypop7.optimizers.core.optimizer import Optimizer
from pypop7.optimizers.rs.prs import PRS


class SRS(PRS):
    """Simple Random Search (SRS).

    .. note:: `SRS` is an *adaptive* random search method, originally designed by Rosenstein and `Barto
       <https://people.cs.umass.edu/~barto/>`_ for **direct policy search** from reinforcement learning.
       Since it uses the simple *individual-based* random sampling strategy, it easily suffers from a *limited*
       exploration ability for large-scale black-box optimization (LSBBO). Therefore, it is **highly recommended**
       to first attempt other more advanced (e.g. population-based) methods for LSBBO.

       Here we include it only for *benchmarking* purpose.

    Parameters
    ----------
    problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`);
              and with the following particular settings (`keys`):
                * 'x'         - initial (starting) point (`array_like`),
                * 'sigma'     - initial (global) step-size (`float`),
                * 'alpha'     - factor of (global) step-size (`float`, default: `0.3`),
                * 'beta'      - adjustment probability for exploration-exploitation trade-off (`float`, default: `0`),
                * 'gamma'     - factor of search decay (`float`, default: `0.99`),
                * 'min_sigma' - minimum of (global) step-size (`float`, default: `0.01`).

    Examples
    --------
    Use the Random Search optimizer `SRS` to minimize the well-known test function
    `Rosenbrock <http://en.wikipedia.org/wiki/Rosenbrock_function>`_:

    .. code-block:: python
       :linenos:

       >>> import numpy
       >>> from pypop7.benchmarks.base_functions import rosenbrock  # function to be minimized
       >>> from pypop7.optimizers.rs.srs import SRS
       >>> problem = {'fitness_function': rosenbrock,  # define problem arguments
       ...            'ndim_problem': 2,
       ...            'lower_boundary': -5 * numpy.ones((2,)),
       ...            'upper_boundary': 5 * numpy.ones((2,))}
       >>> options = {'max_function_evaluations': 5000,  # set optimizer options
       ...            'seed_rng': 2022,
       ...            'x': 3 * numpy.ones((2,)),
       ...            'sigma': 0.1}
       >>> srs = SRS(problem, options)  # initialize the optimizer class
       >>> results = srs.optimize()  # run the optimization process
       >>> # return the number of function evaluations and best-so-far fitness
       >>> print(f"SRS: {results['n_function_evaluations']}, {results['best_so_far_y']}")
       SRS: 5000, 0.0017821578376762473

    Attributes
    ----------
    x                           : `array_like`
                                  initial (starting) point.
    sigma                       : `float`
                                  (global) step-size.
    alpha                       : `float`
                                  factor of (global) step-size.
    beta                        : `float`
                                  adjustment probability for exploration-exploitation trade-off.
    gamma                       : `float`
                                  factor of search decay.
    min_sigma                   :  `float`
                                  minimum of (global) step-size.

    References
    ----------
    Rosenstein, M.T. and Grupen, R.A., 2002, May.
    Velocity-dependent dynamic manipulability.
    In Proceedings of IEEE International Conference on Robotics and Automation (pp. 2424-2429). IEEE.
    https://ieeexplore.ieee.org/abstract/document/1013595

    Rosenstein, M.T. and Barto, A.G., 2001, August.
    Robot weightlifting by direct policy search.
    In International Joint Conference on Artificial Intelligence (pp. 839-846).
    https://dl.acm.org/doi/abs/10.5555/1642194.1642206
    """
    def __init__(self, problem, options):
        # only support normally-distributed random sampling during optimization
        options['sampling_distribution'] = 0
        PRS.__init__(self, problem, options)
        self.alpha = options.get('alpha', 0.3)  # factor of (global) step-size
        assert self.alpha > 0.0, f'`self.alpha` == {self.alpha}, but should > 0.0.'
        self.beta = options.get('beta', 0.0)  # adjustment probability for exploration-exploitation trade-off
        assert 0.0 <= self.beta <= 1.0, f'`self.beta` == {self.beta}, but should >= 0.0 and <= 1.0.'
        self.gamma = options.get('gamma', 0.99)  # factor of search decay
        assert 0.0 <= self.gamma <= 1.0, f'`self.gamma` == {self.gamma}, but should >= 0.0 and <= 1.0.'
        self.min_sigma = options.get('min_sigma', 0.01)  # minimum of (global) step-size
        assert self.min_sigma > 0.0, f'`self.min_sigma` == {self.min_sigma}, but should > 0.0.'

    def initialize(self, args=None):
        if self.x is None:
            x = self.rng_initialization.uniform(self.initial_lower_boundary, self.initial_upper_boundary)
        else:
            x = np.copy(self.x)
        y = self._evaluate_fitness(x, args)
        return x, y

    def iterate(self, x=None, args=None):
        delta_x = self.sigma*self.rng_optimization.standard_normal(size=(self.ndim_problem,))
        y = self._evaluate_fitness(x + delta_x, args)  # random perturbation
        if self.rng_optimization.uniform() < self.beta:
            x += self.alpha*delta_x
        else:
            x += self.alpha*(self.best_so_far_x - x)
        self._n_generations += 1
        return x, y

    def optimize(self, fitness_function=None, args=None):
        fitness = Optimizer.optimize(self, fitness_function)
        x, y = self.initialize(args)
        if self.saving_fitness:
            fitness.append(y)
        self._print_verbose_info(y)
        while True:
            x, y = self.iterate(x, args)
            if self.saving_fitness:
                fitness.append(y)
            self._print_verbose_info(y)
            if self._check_terminations():
                break
            self.sigma = np.maximum(self.gamma*self.sigma, self.min_sigma)
        return self._collect_results(fitness)



import numpy as np

from pypop7.optimizers.core.optimizer import Optimizer
from pypop7.optimizers.rs.rs import RS


class BES(RS):
    """BErnoulli Smoothing (BES).

    References
    ----------
    Gao, K. and Sener, O., 2022, June.
    Generalizing Gaussian Smoothing for Random Search.
    In International Conference on Machine Learning (pp. 7077-7101). PMLR.
    https://proceedings.mlr.press/v162/gao22f.html
    https://icml.cc/media/icml-2022/Slides/16434.pdf
    """
    def __init__(self, problem, options):
        RS.__init__(self, problem, options)
        self.n_individuals = options.get('n_individuals', 100)  # number of individuals (samples)
        self.lr = options.get('lr', 0.001)  # learning rate
        self.c = options.get('c', 0.1)  # factor of finite-difference gradient estimate
        self.verbose = options.get('verbose', 10)

    def initialize(self, is_restart=False):
        if is_restart or (self.x is None):
            x = self.rng_initialization.uniform(self.initial_lower_boundary, self.initial_upper_boundary)
        else:
            x = np.copy(self.x)
        return x

    def iterate(self, args=None, x=None, fitness=None):  # for each iteration (generation)
        gradient = np.zeros((self.ndim_problem,))  # estimated gradient
        y_base = self._evaluate_fitness(x, args)  # for finite-difference gradient estimate
        if self.saving_fitness:
            fitness.append(y_base)
        self._print_verbose_info(y_base)
        y = np.empty((self.n_individuals,))
        for i in range(self.n_individuals):
            if self._check_terminations():
                return x
            # set directional gradient based on Bernoulli distribution
            dg = self.rng_optimization.binomial(n=1, p=0.5, size=(self.ndim_problem,))
            dg = (dg - 0.5)/0.5
            y[i] = self._evaluate_fitness(x + self.c*dg, args)
            gradient += (y[i] - y_base)*dg
            if self.saving_fitness:
                fitness.append(y[i])
        gradient /= (self.c*self.n_individuals)
        x -= self.lr*gradient  # stochastic gradient descent (SGD)
        self._n_generations += 1
        self._print_verbose_info(y)
        return x

    def optimize(self, fitness_function=None, args=None):  # for all iterations (generations)
        fitness = Optimizer.optimize(self, fitness_function)
        x = self.initialize()
        while not self._check_terminations():
            x = self.iterate(args, x, fitness)
        return self._collect_results(fitness)



"""Repeat DifferentialEvolution algorithm in Scipy library
    The code is as following:
    import numpy as np
    import time
    from scipy.optimize import differential_evolution

    def sphere(x):
        y = np.sum(np.power(x, 2))
        return y

    start_time = time.time()
    ndim_problem = 30
    result = differential_evolution(sphere, bounds=[(-100, 100)] * ndim_problem, strategy='best1bin',
                                popsize=100, maxiter=800, disp=True)
    print("result x:", result.x, "result f:", result.fun)
    print("Runtime: {:7.5e}".format(time.time() - start_time))

    Compare pypop's JADE algorithm with it:
    scipy result:
    best: 2.11193e-24

    pypop result:
    best: 2.1046682114162533e-31
"""
import numpy as np

from pypop7.benchmarks.base_functions import sphere
from pypop7.optimizers.de.jade import JADE

if __name__ == '__main__':
    ndim_problem = 30
    problem = {'fitness_function': sphere,
               'ndim_problem': ndim_problem,
               'lower_boundary': -100 * np.ones((ndim_problem,)),
               'upper_boundary': 100 * np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 800 * 100,
               'n_individuals': 100,
               'seed_rng': 0,  # undefined in the original paper
               }
    jade = JADE(problem, options)
    results = jade.optimize()
    print(results)
    print(results['best_so_far_y'])



import numpy as np

from pypop7.optimizers.es.es import ES


class OPOC(ES):
    """(1+1)-Cholesky-CMA-ES (OPOC).

    Reference
    ---------
    Igel, C., Suttorp, T. and Hansen, N., 2006, July.
    A computational efficient covariance matrix update and a (1+1)-CMA for evolution strategies.
    In Proceedings of Annual Conference on Genetic and Evolutionary Computation (pp. 453-460). ACM.
    https://dl.acm.org/doi/abs/10.1145/1143997.1144082
    (See Algorithm 2 for details.)
    """
    def __init__(self, problem, options):
        options['n_individuals'] = 1  # mandatory setting for OPOC
        options['n_parents'] = 1  # mandatory setting for OPOC
        ES.__init__(self, problem, options)
        if self.eta_sigma is None:
            self.eta_sigma = 1 / (1 + self.ndim_problem / 2)
        self.p_ts = options.get('p_ts', 2 / 11)
        self.c_p = options.get('c_p', 1 / 12)
        self.c_c = options.get('c_c', 2 / (self.ndim_problem + 2))
        self.c_cov = options.get('c_cov', 2 / (np.power(self.ndim_problem, 2) + 6))
        self.p_t = options.get('p_t', 0.44)
        self.c_a = options.get('c_a', np.sqrt(1 - self.c_cov))

    def initialize(self, args=None, is_restart=False):
        mean = self._initialize_mean(is_restart)  # mean of Gaussian search distribution
        y = self._evaluate_fitness(mean, args)  # fitness
        a = np.diag(np.ones(self.ndim_problem,))  # linear transformation (Cholesky factors)
        best_so_far_y, p_s = np.copy(y), self.p_ts
        return mean, y, a, best_so_far_y, p_s

    def iterate(self, args=None, mean=None, a=None, best_so_far_y=None, p_s=None):
        # sample and evaluate (only one) offspring
        z = self.rng_optimization.standard_normal((self.ndim_problem,))
        x = mean + self.sigma * np.dot(a, z)
        y = self._evaluate_fitness(x, args)
        if y <= best_so_far_y:
            l_s = 1
        else:
            l_s = 0
        p_s = (1 - self.c_p) * p_s + self.c_p * l_s
        self.sigma *= np.exp(self.eta_sigma * (p_s - self.p_ts) / (1 - self.p_ts))
        if y <= best_so_far_y:
            mean, best_so_far_y = x, y
            if p_s < self.p_t:
                z_norm, c_a = np.power(np.linalg.norm(z), 2), np.power(self.c_a, 2)
                a = self.c_a * a + self.c_a / z_norm * (np.sqrt(1 + ((1 - c_a) * z_norm) / c_a) - 1) * np.dot(
                    np.dot(a, z[:, np.newaxis]), z[np.newaxis, :])
        return mean, y, a, best_so_far_y, p_s

    def restart_initialize(self, args=None, mean=None, y=None,
                           a=None, best_so_far_y=None, p_s=None, fitness=None):
        self._fitness_list.append(self.best_so_far_y)
        is_restart_1, is_restart_2 = self.sigma < self.sigma_threshold, False
        if len(self._fitness_list) >= self.stagnation:
            is_restart_2 = (self._fitness_list[-self.stagnation] - self._fitness_list[-1]) < self.fitness_diff
        is_restart = bool(is_restart_1) or bool(is_restart_2)
        if is_restart:
            self.n_restart += 1
            self.sigma = np.copy(self._sigma_bak)
            mean, y, a, best_so_far_y, p_s = self.initialize(args, is_restart)
            fitness.append(y)
            self._fitness_list = [best_so_far_y]
        return mean, y, a, best_so_far_y, p_s

    def optimize(self, fitness_function=None, args=None):  # for all generations (iterations)
        fitness = ES.optimize(self, fitness_function)
        mean, y, a, best_so_far_y, p_s = self.initialize(args)
        fitness.append(y)
        while True:
            mean, y, a, best_so_far_y, p_s = self.iterate(args, mean, a, best_so_far_y, p_s)
            if self.record_fitness:
                fitness.append(y)
            if self._check_terminations():
                break
            self._n_generations += 1
            self._print_verbose_info(y)
            if self.is_restart:
                mean, y, a, best_so_far_y, p_s = self.restart_initialize(
                    args, mean, y, a, best_so_far_y, p_s, fitness)
        results = self._collect_results(fitness, mean)
        return results



"""Repeat Table 2 and 3 from the following paper:
    Ros, R. and Hansen, N., 2008, September.
    A simple modification in CMA-ES achieving linear time and space complexity.
    In International Conference on Parallel Problem Solving from Nature (pp. 296-305).
    Springer, Berlin, Heidelberg.
    https://link.springer.com/chapter/10.1007/978-3-540-87700-4_30

    Since our code obtains very similar performance as the original paper, we argue that
    the repeatability of `SEPCMAES` can be well-documented (*at least partly*).
"""
import numpy as np

from pypop7.benchmarks.base_functions import rosenbrock, ellipsoid
from pypop7.optimizers.es.sepcmaes import SEPCMAES


def _different_powers(x):
    y = np.sum(np.power(np.abs(x), 2 + 10 * np.linspace(0, 1, x.size)))
    return y


def _hyper_ellipsoid(x):
    y = 0
    for i in range(x.size):
        y += np.power(x[i] * (i + 1), 2)
    return y


if __name__ == '__main__':
    # Table 2
    ndim_problem = 30
    for f in [_hyper_ellipsoid, _different_powers, rosenbrock]:
        problem = {'fitness_function': f,
                   'ndim_problem': ndim_problem}
        options = {'max_function_evaluations': 200 * 1e3,
                   'fitness_threshold': 1e-6,
                   'seed_rng': 2022,  # not given in the original paper
                   'verbose_frequency': 1000,
                   'record_fitness': True,
                   'record_fitness_frequency': 1000,
                   'is_restart': False}
        if f == _hyper_ellipsoid:
            options['fitness_threshold'] = 1e-10
            options['sigma'] = 1
            options['x'] = np.ones((ndim_problem,))
        elif f == _different_powers:
            options['fitness_threshold'] = 1e-20
            options['sigma'] = 1
            options['x'] = np.ones((ndim_problem,))
        elif f == rosenbrock:
            options['fitness_threshold'] = 1e-6
            options['sigma'] = 0.1
            options['x'] = np.zeros((ndim_problem,))
        solver = SEPCMAES(problem, options)
        results = solver.optimize()
        print(results)
        print(f.__name__, 'dim =', problem['ndim_problem'],
              'fitness threshold =', options['fitness_threshold'],
              'functions evaluations =', results['n_function_evaluations'])
    # _hyper_ellipsoid dim = 30 fitness threshold = 1e-10 functions evaluations = 6261
    # _different_powers dim = 30 fitness threshold = 1e-20 functions evaluations = 10238
    # rosenbrock dim = 30 fitness threshold = 1e-06 functions evaluations = 102590

    # Table 3
    ndim_problem = 20
    for f in [ellipsoid, rosenbrock]:
        problem = {'fitness_function': f,
                   'ndim_problem': ndim_problem}
        options = {'max_function_evaluations': 200 * 1e3,
                   'fitness_threshold': 1e-9,
                   'seed_rng': 2022,  # not given in the original paper
                   'verbose_frequency': 1000,
                   'record_fitness': True,
                   'record_fitness_frequency': 1000,
                   'is_restart': False}
        if f == ellipsoid:
            options['sigma'] = 1
            options['x'] = np.ones((ndim_problem,))
        elif f == rosenbrock:
            options['sigma'] = 0.1
            options['x'] = np.zeros((ndim_problem,))
        solver = SEPCMAES(problem, options)
        results = solver.optimize()
        print(results)
        print(f.__name__, 'dim =', problem['ndim_problem'],
              'fitness threshold =', options['fitness_threshold'],
              'function evaluations =', results['n_function_evaluations'])
    # ellipsoid dim = 20 fitness threshold = 1e-09 function evaluations = 5861
    # rosenbrock dim = 20 fitness threshold = 1e-09 function evaluations = 110151



"""Repeat StochasticHillClimber algorithm in pybrain with python version=2.7
    The code is as following:
    import time
    import numpy as np
    from pybrain.optimization.hillclimber import StochasticHillClimber as SHC

    def ellipsoid(x):
        x = np.power(x, 2)
        y = np.dot(np.power(10, 6 * np.linspace(0, 1, x.size)), x)
        return y

    solver = SHC(ellipsoid, 4 * np.ones((1000,)), minimize=True, maxEvaluations=2e6, verbose=True)
    start_time = time.time()
    solver.learn()
    print("Runtime: {:7.5e}".format(time.time() - start_time))

    pybrain result:
    best: 12968919.810064487

    pypop result:
    best: 10389166.423976313
"""
import numpy as np
import time

from pypop7.benchmarks.base_functions import ellipsoid
from pypop7.optimizers.rs.arhc import ARHC


if __name__ == '__main__':
    start_run = time.time()
    ndim_problem = 1000
    problem = {'fitness_function': ellipsoid,
               'ndim_problem': ndim_problem,
               'upper_boundary': 5.0 * np.ones((ndim_problem,)),
               'lower_boundary': -5.0 * np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 2e6,
               'fitness_threshold': 1e-10,
               'seed_rng': 0,
               'x': 4 * np.ones((ndim_problem,)),
               'sigma': 0.1,
               'temperature': 100,
               'verbose': 200000,
               'saving_fitness': 200000}
    arhc = ARHC(problem, options)
    results = arhc.optimize()
    print(results)
    print('*** Runtime: {:7.5e}'.format(time.time() - start_run))


"""Repeat Fig. 44.2a from the following paper for `RES`:
    Hansen, N., Arnold, D.V. and Auger, A., 2015.
    Evolution strategies.
    In Springer Handbook of Computational Intelligence (pp. 871-898).
    Springer, Berlin, Heidelberg.
    https://link.springer.com/chapter/10.1007%2F978-3-662-43505-2_44
"""
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns

from pypop7.benchmarks.base_functions import sphere
from pypop7.optimizers.es.res import RES


if __name__ == '__main__':
    sns.set_theme(style='darkgrid')
    plt.figure()
    ndim = 10
    for i in range(3):
        problem = {'fitness_function': sphere,
                   'ndim_problem': ndim}
        options = {'max_function_evaluations': 1500,
                   'seed_rng': i,  # undefined in the original paper
                   'saving_fitness': 1,
                   'x': np.ones((ndim,)),
                   'sigma': 1e-9,
                   'lr_sigma': 1.0/(1.0 + 10.0/3.0),
                   'is_restart': False}
        res = RES(problem, options)
        results = res.optimize()
        fitness = results['fitness']
        plt.plot(fitness[:, 0], np.sqrt(fitness[:, 1]), 'b')
    plt.xticks([0, 500, 1000, 1500])
    plt.xlim([0, 1500])
    plt.yscale('log')
    plt.yticks([1e-9, 1e-6, 1e-3, 1e0])
    plt.show()



import numpy as np

from pypop7.optimizers.de.de import DE


class TDE(DE):
    """Trigonometric-mutation Differential Evolution (TDE).

    Parameters
    ----------
    problem : `dict`
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : `dict`
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`);
              and with the following particular settings (`keys`):
                * 'n_individuals' - population size (`int`, default: `30`),
                * 'f'             - mutation factor (`float`, default: `0.99`),
                * 'cr'            - crossover probability (`float`, default: `0.85`),
                * 'mt'            - trigonometric mutation probability (`float`, default: `0.05`).

    Examples
    --------
    Use the ES optimizer `TDE` to minimize the well-known test function
    `Rosenbrock <http://en.wikipedia.org/wiki/Rosenbrock_function>`_:

    .. code-block:: python
       :linenos:

       >>> import numpy
       >>> from pypop7.benchmarks.base_functions import rosenbrock  # function to be minimized
       >>> from pypop7.optimizers.de.tde import TDE
       >>> problem = {'fitness_function': rosenbrock,  # define problem arguments
       ...            'ndim_problem': 2,
       ...            'lower_boundary': -5 * numpy.ones((2,)),
       ...            'upper_boundary': 5 * numpy.ones((2,))}
       >>> options = {'max_function_evaluations': 5000,  # set optimizer options
       ...            'seed_rng': 0}
       >>> tde = TDE(problem, options)  # initialize the optimizer class
       >>> results = tde.optimize()  # run the optimization process
       >>> # return the number of function evaluations and best-so-far fitness
       >>> print(f"TDE: {results['n_function_evaluations']}, {results['best_so_far_y']}")
       TDE: 5000, 6.420787226215637e-21

    For its correctness checking of coding, refer to `this code-based repeatability report
    <https://tinyurl.com/56frubv2>`_ for more details.

    Attributes
    ----------
    n_individuals : `int`
                    number of offspring, offspring population size.
    f             : `float`
                    mutation factor.
    cr            : `float`
                    crossover probability.
    mt            : 'float
                    trigonometric mutation probability

    References
    ----------
    Fan, H.Y. and Lampinen, J., 2003.
    A trigonometric mutation operation to differential evolution.
    Journal of Global Optimization, 27(1), pp.105-129.
    https://link.springer.com/article/10.1023/A:1024653025686
    """
    def __init__(self, problem, options):
        DE.__init__(self, problem, options)
        self.n_individuals = options.get('n_individuals', 30)  # population size
        self.f = options.get('f', 0.99)  # mutation factor
        self.mt = options.get('mt', 0.05)  # trigonometric mutation probability
        self.cr = options.get('cr', 0.85)  # crossover probability

    def initialize(self, args=None):
        x = self.rng_initialization.uniform(self.initial_lower_boundary, self.initial_upper_boundary,
                                            size=(self.n_individuals, self.ndim_problem))  # population
        y = np.empty((self.n_individuals,))  # fitness
        for i in range(self.n_individuals):
            if self._check_terminations():
                break
            y[i] = self._evaluate_fitness(x[i], args)
        return x, y

    def mutate(self, x=None, y=None):
        v, base = np.empty((self.n_individuals, self.ndim_problem)), np.arange(self.n_individuals)
        for i in range(self.n_individuals):
            if self.rng_optimization.random() < self.mt:  # trigonometric mutation
                r0 = self.rng_optimization.choice(np.setdiff1d(base, i))
                r1 = self.rng_optimization.choice(np.setdiff1d(base, [i, r0]))
                r2 = self.rng_optimization.choice(np.setdiff1d(base, [i, r0, r1]))
                p = np.abs(y[r0]) + np.abs(y[r1]) + np.abs(y[r2])
                p1, p2, p3 = np.abs(y[r0])/p, np.abs(y[r1])/p, np.abs(y[r2])/p
                v[i] = ((x[r0] + x[r1] + x[r2])/3.0 + (p2 - p1)*(x[r0] - x[r1]) +
                        (p3 - p2)*(x[r1] - x[r2]) + (p1 - p3)*(x[r2] - x[r0]))
            else:  # same as the original ES version (`DE/rand/1/bin`)
                r0 = self.rng_optimization.choice(np.setdiff1d(base, i))
                r1 = self.rng_optimization.choice(np.setdiff1d(base, [i, r0]))
                r2 = self.rng_optimization.choice(np.setdiff1d(base, [i, r0, r1]))
                v[i] = x[r0] + self.f*(x[r1] - x[r2])
        return v

    def crossover(self, v=None, x=None):
        for i in range(self.n_individuals):
            k = self.rng_optimization.integers(self.ndim_problem)
            for j in range(self.ndim_problem):
                if (self.rng_optimization.random() <= self.cr) or (k == j):
                    continue
                else:
                    v[i][j] = x[i][j]
        return v

    def select(self, v=None, x=None, y=None, args=None):
        for i in range(self.n_individuals):
            if self._check_terminations():
                break
            yy = self._evaluate_fitness(v[i], args)
            if yy <= y[i]:
                x[i], y[i] = v[i], yy
        return x, y

    def iterate(self, args=None, x=None, y=None):
        return self.select(self.crossover(self.mutate(x, y), x), x, y, args)

    def optimize(self, fitness_function=None, args=None):
        fitness = DE.optimize(self, fitness_function)
        x, y = self.initialize(args)
        if self.saving_fitness:
            fitness.extend(y)
        self._print_verbose_info(y)
        while True:
            x, y = self.iterate(args, x, y)
            if self.saving_fitness:
                fitness.extend(y)
            if self._check_terminations():
                break
            self._n_generations += 1
            self._print_verbose_info(y)
        return self._collect_results(fitness)



import numpy as np

from pypop7.optimizers.es.es import ES
from pypop7.optimizers.es.dsaes import DSAES


class CSAES(DSAES):
    """Cumulative Step-size Adaptation Evolution Strategy (CSAES).

    .. note:: `CSAES` adapts the *individual* step-sizes on-the-fly with *small* populations, according to the
       well-known `Cumulative Step-size Adaptation (CSA) <http://link.springer.com/chapter/10.1007/3-540-58484-6_263>`_
       rule from the evolutionary computation community.

       AKA cumulative path length control.

    Parameters
    ----------
    problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`);
              and with the following particular settings (`keys`):
                * 'mean'          - initial (starting) point, mean of Gaussian search distribution (`array_like`),
                * 'sigma'         - initial global step-size (σ), mutation strength (`float`),
                * 'n_individuals' - number of offspring (λ: lambda), offspring population size (`int`, default:
                  `4 + int(np.floor(3*np.log(problem.get('ndim_problem'))))`),
                * 'n_parents'     - number of parents (μ: mu), parental population size (`int`, default:
                  `int(options['n_individuals']/4)`),
                * 'lr_sigma'     - learning rate of global step-size (`float`, default:
                  `np.sqrt(options['n_parents']/(problem['ndim_problem'] + options['n_parents']))`).

    Examples
    --------
    Use the ES optimizer `CSAES` to minimize the well-known test function
    `Rosenbrock <http://en.wikipedia.org/wiki/Rosenbrock_function>`_:

    .. code-block:: python
       :linenos:

       >>> import numpy
       >>> from pypop7.benchmarks.base_functions import rosenbrock  # function to be minimized
       >>> from pypop7.optimizers.es.csaes import CSAES
       >>> problem = {'fitness_function': rosenbrock,  # define problem arguments
       ...            'ndim_problem': 2,
       ...            'lower_boundary': -5 * numpy.ones((2,)),
       ...            'upper_boundary': 5 * numpy.ones((2,))}
       >>> options = {'max_function_evaluations': 5000,  # set optimizer options
       ...            'seed_rng': 2022,
       ...            'mean': 3 * numpy.ones((2,)),
       ...            'sigma': 0.1}
       >>> csaes = CSAES(problem, options)  # initialize the optimizer class
       >>> results = csaes.optimize()  # run the optimization process
       >>> # return the number of function evaluations and best-so-far fitness
       >>> print(f"CSAES: {results['n_function_evaluations']}, {results['best_so_far_y']}")
       CSAES: 5000, 0.026416618757717083

    Attributes
    ----------
    n_individuals : `int`
                    number of offspring (λ: lambda), offspring population size.
    n_parents     : `int`
                    number of parents (μ: mu), parental population size.
    mean          : `array_like`
                    mean of Gaussian search distribution.
    sigma         : `float`
                    mutation strength.
    lr_sigma      : `float`
                    learning rate of global step-size.

    References
    ----------
    Hansen, N., Arnold, D.V. and Auger, A., 2015.
    Evolution strategies.
    In Springer Handbook of Computational Intelligence (pp. 871-898). Springer, Berlin, Heidelberg.
    https://link.springer.com/chapter/10.1007%2F978-3-662-43505-2_44
    (See Algorithm 44.6 for details.)

    Ostermeier, A., Gawelczyk, A. and Hansen, N., 1994, October.
    Step-size adaptation based on non-local use of selection information
    In International Conference on Parallel Problem Solving from Nature (pp. 189-198).
    Springer, Berlin, Heidelberg.
    http://link.springer.com/chapter/10.1007/3-540-58484-6_263
    """
    def __init__(self, problem, options):
        if options.get('n_individuals') is None:
            options['n_individuals'] = 4 + int(np.floor(3*np.log(problem.get('ndim_problem'))))
        if options.get('n_parents') is None:
            options['n_parents'] = int(options['n_individuals']/4)
        if options.get('lr_sigma') is None:
            options['lr_sigma'] = np.sqrt(options['n_parents']/(
                    problem['ndim_problem'] + options['n_parents']))
        DSAES.__init__(self, problem, options)
        self._s_1 = None  # for Line 8
        self._s_2 = None  # for Line 8
        # E[||N(0,I)||]: expectation of chi distribution
        self._e_chi = np.sqrt(self.ndim_problem)*(
                1.0 - 1.0/(4.0*self.ndim_problem) + 1.0/(21.0*np.power(self.ndim_problem, 2)))

    def initialize(self, is_restart=False):
        self._s_1 = 1.0 - self.lr_sigma
        self._s_2 = np.sqrt(self.lr_sigma*(2.0 - self.lr_sigma)*self.n_parents)
        self._axis_sigmas = self._sigma_bak*np.ones((self.ndim_problem,))
        z = np.empty((self.n_individuals, self.ndim_problem))  # noise for offspring population
        x = np.empty((self.n_individuals, self.ndim_problem))  # offspring population
        mean = self._initialize_mean(is_restart)  # mean of Gaussian search distribution
        s = np.zeros((self.ndim_problem,))  # evolution path
        y = np.empty((self.n_individuals,))  # fitness (no evaluation)
        return z, x, mean, s, y

    def iterate(self, z=None, x=None, mean=None, y=None, args=None):
        # sample offspring population (Line 4)
        for k in range(self.n_individuals):
            if self._check_terminations():
                return z, x, y
            z[k] = self.rng_optimization.standard_normal((self.ndim_problem,))  # Line 5
            x[k] = mean + self._axis_sigmas*z[k]  # Line 6
            y[k] = self._evaluate_fitness(x[k], args)
        return z, x, y

    def _update_distribution(self, z=None, x=None, s=None, y=None):
        order = np.argsort(y)[:self.n_parents]  # Line 7
        s = self._s_1*s + self._s_2*np.mean(z[order], axis=0)  # Line 8
        sigmas_1 = np.power(np.exp(np.abs(s)/self._e_hnd - 1.0), 1.0/(3.0*self.ndim_problem))
        sigmas_2 = np.power(np.exp(np.linalg.norm(s)/self._e_chi - 1.0),
                            self.lr_sigma/(1.0 + np.sqrt(self.n_parents/self.ndim_problem)))
        self._axis_sigmas *= (sigmas_1*sigmas_2)  # Line 9
        mean = np.mean(x[order], axis=0)  # Line 11
        return s, mean

    def restart_reinitialize(self, z=None, x=None, mean=None, s=None, y=None):
        if self._restart_reinitialize():
            self.n_parents = int(self.n_individuals/4)
            self.lr_sigma = np.sqrt(self.n_parents/(self.ndim_problem + self.n_parents))
            z, x, mean, s, y = self.initialize(True)
        return z, x, mean, s, y

    def optimize(self, fitness_function=None, args=None):  # for all generations (iterations)
        fitness = ES.optimize(self, fitness_function)
        z, x, mean, s, y = self.initialize()
        while True:
            # sample and evaluate offspring population
            z, x, y = self.iterate(z, x, mean, y, args)
            if self.saving_fitness:
                fitness.extend(y)
            if self._check_terminations():
                break
            s, mean = self._update_distribution(z, x, s, y)
            self._print_verbose_info(y)
            self._n_generations += 1
            if self.is_restart:
                z, x, mean, s, y = self.restart_reinitialize(z, x, mean, s, y)
        results = self._collect_results(fitness, mean)
        results['s'] = s
        results['_axis_sigmas'] = self._axis_sigmas
        return results



from pypop7.optimizers.pso.pso import PSO
from pypop7.optimizers.pso.spso import SPSO
from pypop7.optimizers.pso.spsol import SPSOL
from pypop7.optimizers.pso.clpso import CLPSO
from pypop7.optimizers.pso.ccpso2 import CCPSO2


__all__ = [PSO,  # base (abstract) class
           SPSO, SPSOL,  # standard and canonical versions
           CLPSO,  # competitors for medium- and small-scale black-box problems
           CCPSO2]  # variants for large-scale black-box optimization



import numpy as np

from pypop7.optimizers.ga.ga import GA
from athena.active import ActiveSubspaces


class ASGA(GA):
    """Active Subspaces extension of the standard GA (ASGA)

    References
    ----------
    Repeat the following paper for `ASGA`:
    Demo, N., Tezzele, M. and Rozza, G., 2021.
    A supervised learning approach involving active subspaces for an efficient genetic algorithm in high-dimensional
    optimization problems.
    SIAM Journal on Scientific Computing, 43(3), pp.B831-B853.
    https://epubs.siam.org/doi/10.1137/20M1345219
    """
    def __init__(self, problem, options):
        GA.__init__(self, problem, options)
        self.crossover_prob = options.get('crossover_prob', 0.5)  # crossover probability
        self.mutate_prob = options.get('mutate_prob', 0.5)  # mutate probability
        self.ndim_subspace = options.get('ndim_subspace', 1)  # the number of active dimensions of subspace
        self.n_b = options.get('b', 2)  # the number of back-mapped points is 2
        self.n_individuals_subspace = int(self.n_individuals / self.n_b)
        self.alpha = options.get('alpha', 1)
        self.n_initial_individuals = options.get('n_initial_individuals', 2000)
        self.n_individuals = options.get('n_individuals', 200)

    def initialize(self):
        x = self.rng_initialization.uniform(self.initial_lower_boundary, self.initial_upper_boundary,
                                            size=(self.n_initial_individuals, self.ndim_problem))  # initial population
        y = np.empty((self.n_initial_individuals,))  # fitness
        for i in range(self.n_initial_individuals):
            if self._check_terminations():
                break
            y[i] = self._evaluate_fitness(x[i])
        self._n_generations += 1
        x_as, y_as = np.copy(x), np.copy(y)
        return x, y, x_as, y_as

    def build_active_space(self, x_as=None, y_as=None):
        active_subspace = ActiveSubspaces(dim=self.ndim_subspace, method='local', n_boot=100)
        active_subspace.fit(inputs=x_as, outputs=y_as)
        return active_subspace

    def select(self, x=None, y=None):
        return x[np.argsort(y)[:self.n_individuals_subspace]]

    def crossover(self, x=None):
        xx = np.copy(x)
        for i in range(self.n_individuals_subspace):
            x1, x2 = self.rng_optimization.choice(x, (2,))
            for j in range(self.ndim_subspace):
                if self.rng_optimization.random() < self.crossover_prob:
                    r = self.rng_optimization.uniform(-1 * self.alpha, 1 + self.alpha)
                    xx[i][j] = (1-r)*x1[j] + r*x2[j]
        return xx

    def mutate(self, x=None):
        for i in range(self.n_individuals_subspace):
            for j in range(self.ndim_subspace):
                if self.rng_optimization.random() < self.mutate_prob:
                    x[i][j] *= (1 + self.rng_optimization.normal(0, 0.1))
        return x

    def iterate(self, x=None, y=None, x_as=None, y_as=None, args=None):
        active_subspace = self.build_active_space(x_as, y_as)  # build active space
        xx = self.select(x, y)  # select
        xx = active_subspace.transform(xx)[0]  # forward (reduction)
        xx = self.crossover(xx)  # crossover (mate)
        xx = self.mutate(xx)  # mutate
        x = active_subspace.inverse_transform(xx, self.n_b)[0]  # backward
        y = np.empty((self.n_individuals,))
        for i in range(self.n_individuals):
            y[i] = self._evaluate_fitness(x[i], args)
        x_as = np.vstack((x_as, x))
        y_as = np.hstack((y_as, y))
        return x, y, x_as, y_as

    def optimize(self, fitness_function=None):
        fitness = GA.optimize(self, fitness_function)
        x, y, x_as, y_as = self.initialize()
        fitness.extend(y)
        while True:
            x, y, x_as, y_as = self.iterate(x, y, x_as, y_as)
            if self.saving_fitness:
                fitness.extend(y)
            if self._check_terminations():
                break
            self._n_generations += 1
            self._print_verbose_info(y)
        return self._collect_results(fitness)


import numpy as np
from scipy.stats import norm

from pypop7.optimizers.es.es import ES


class MMES(ES):
    """Mixture Model-based Evolution Strategy (MMES).

    Parameters
    ----------
    problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime to be allowed (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`);
              and with the following particular settings (`keys`):
                * 'sigma'         - initial global step-size, aka mutation strength (`float`),
                * 'mean'          - initial (starting) point, aka mean of Gaussian search distribution (`array_like`),

                  * if not given, it will draw a random sample from the uniform distribution whose search range is
                    bounded by `problem['lower_boundary']` and `problem['upper_boundary']`.

                * 'm'             - number of candidate direction vectors (`int`, default:
                  `2*int(np.ceil(np.sqrt(self.ndim_problem)))`),
                * 'c_c'           - learning rate of evolution path update (`float`, default:
                  `0.4/np.sqrt(self.ndim_problem)`),
                * 'ms'            - mixing strength (`int`, default: `4`),
                * 'c_s'           - learning rate of global step-size adaptation (`float`, default: `0.3`),
                * 'a_z'           - target significance level (`float`, default: `0.05`),
                * 'distance'      - minimal distance of updating evolution paths (`int`, default:
                  `int(np.ceil(1.0/self.c_c))`),
                * 'n_individuals' - number of offspring, aka offspring population size (`int`, default:
                  `4 + int(3*np.log(self.ndim_problem))`),
                * 'n_parents'     - number of parents, aka parental population size (`int`, default:
                  `int(self.n_individuals/2)`).

    Examples
    --------
    Use the `ES` optimizer `MMES` to minimize the well-known test function
    `Rosenbrock <http://en.wikipedia.org/wiki/Rosenbrock_function>`_:

    .. code-block:: python
       :linenos:

       >>> import numpy
       >>> from pypop7.benchmarks.base_functions import rosenbrock  # function to be minimized
       >>> from pypop7.optimizers.es.mmes import MMES
       >>> problem = {'fitness_function': rosenbrock,  # define problem arguments
       ...            'ndim_problem': 200,
       ...            'lower_boundary': -5*numpy.ones((200,)),
       ...            'upper_boundary': 5*numpy.ones((200,))}
       >>> options = {'max_function_evaluations': 500000,  # set optimizer options
       ...            'seed_rng': 2022,
       ...            'mean': 3*numpy.ones((200,)),
       ...            'sigma': 0.1}  # the global step-size may need to be tuned for better performance
       >>> mmes = MMES(problem, options)  # initialize the optimizer class
       >>> results = mmes.optimize()  # run the optimization process
       >>> # return the number of function evaluations and best-so-far fitness
       >>> print(f"MMES: {results['n_function_evaluations']}, {results['best_so_far_y']}")
       MMES: 500000, 7.350414979806526

    For its correctness checking of coding, refer to `this code-based repeatability report
    <https://tinyurl.com/3ym72w5m>`_ for more details.

    Attributes
    ----------
    a_z           : `float`
                    target significance level.
    c_c           : `float`
                    learning rate of evolution path update.
    c_s           : `float`
                    learning rate of global step-size adaptation.
    distance      : `int`
                    minimal distance of updating evolution paths.
    m             : `int`
                    number of candidate direction vectors.
    mean          : `array_like`
                    initial mean of Gaussian search distribution.
    ms            : `int`
                    mixing strength.
    n_individuals : `int`
                    number of offspring (λ: lambda), aka offspring population size.
    n_parents     : `int`
                    number of parents (μ: mu), aka parental population size.
    sigma         : `float`
                    final mutation strength.

    References
    ----------
    He, X., Zheng, Z. and Zhou, Y., 2021.
    MMES: Mixture model-based evolution strategy for large-scale optimization.
    IEEE Transactions on Evolutionary Computation, 25(2), pp.320-333.
    https://ieeexplore.ieee.org/abstract/document/9244595

    See the official Matlab version from He:
    https://github.com/hxyokokok/MMES
    """
    def __init__(self, problem, options):
        ES.__init__(self, problem, options)
        # set number of candidate direction vectors
        self.m = options.get('m', 2*int(np.ceil(np.sqrt(self.ndim_problem))))
        # set learning rate of evolution path
        self.c_c = options.get('c_c', 0.4/np.sqrt(self.ndim_problem))
        self.ms = options.get('ms', 4)  # mixing strength (l)
        # set for paired test adaptation (PTA)
        self.c_s = options.get('c_s', 0.3)  # learning rate of global step-size adaptation
        self.a_z = options.get('a_z', 0.05)  # target significance level
        # set minimal distance of updating evolution paths (T)
        self.distance = options.get('distance', int(np.ceil(1.0/self.c_c)))
        # set success probability of geometric distribution (different from 4/n in the original paper)
        self.c_a = options.get('c_a', 3.8/self.ndim_problem)  # same as Matlab code
        self.gamma = options.get('gamma', 1.0 - np.power(1.0 - self.c_a, self.m))
        self._n_mirror_sampling = None
        self._z_1 = np.sqrt(1.0 - self.gamma)
        self._z_2 = np.sqrt(self.gamma/self.ms)
        self._p_1 = 1.0 - self.c_c
        self._p_2 = np.sqrt(self.c_c*(2.0 - self.c_c))
        self._w_1 = 1.0 - self.c_s
        self._w_2 = np.sqrt(self.c_s*(2.0 - self.c_s))

    def initialize(self, args=None, is_restart=False):
        self._n_mirror_sampling = int(np.ceil(self.n_individuals/2))
        x = np.zeros((self.n_individuals, self.ndim_problem))  # offspring population
        mean = self._initialize_mean(is_restart)  # mean of Gaussian search distribution
        p = np.zeros((self.ndim_problem,))  # evolution path (Line 2 in Algorithm 1)
        w = 0.0  # Line 3
        q = np.zeros((self.m, self.ndim_problem))  # candidate direction vectors (Line 4)
        t = np.zeros((self.m,))  # Line 5 (generations recorded)
        v = np.arange(self.m)  # Line 6 (indexes to evolution paths)
        y = np.tile(self._evaluate_fitness(mean, args), (self.n_individuals,))  # fitness
        return x, mean, p, w, q, t, v, y

    def iterate(self, x=None, mean=None, q=None, v=None, y=None, args=None):
        for k in range(self._n_mirror_sampling):  # mirror sampling
            zq = np.zeros((self.ndim_problem,))
            for _ in range(self.ms):
                j_k = v[(self.m - self.rng_optimization.geometric(self.c_a) % self.m) - 1]
                zq += self.rng_optimization.standard_normal()*q[j_k]
            z = self._z_1*self.rng_optimization.standard_normal((self.ndim_problem,))
            z += self._z_2*zq  # Line 13
            x[k] = mean + self.sigma*z  # Line 14
            if (self._n_mirror_sampling + k) < self.n_individuals:
                x[self._n_mirror_sampling + k] = mean - self.sigma*z
        for k in range(self.n_individuals):
            if self._check_terminations():
                return x, y
            y[k] = self._evaluate_fitness(x[k], args)
        return x, y

    def _update_distribution(self, x=None, mean=None, p=None, w=None,
                             q=None, t=None, v=None, y=None, y_bak=None):
        order = np.argsort(y)
        y.sort()  # Line 16
        mean_w = np.zeros((self.ndim_problem,))
        for k in range(self.n_parents):  # Line 17
            mean_w += self._w[k]*x[order[k]]
        p = self._p_1*p + self._p_2*np.sqrt(self._mu_eff)*(mean_w - mean)/self.sigma  # Line 18
        mean = mean_w  # for Line 17
        if self._n_generations < self.m:
            q[self._n_generations] = p
        else:
            k_star = np.argmin(t[v[1:]] - t[v[:(self.m - 1)]])  # Line 19
            k_star += 1
            if t[v[k_star]] - t[v[k_star - 1]] > self.distance:  # Line 20
                k_star = 0  # Line 21
            v = np.append(np.append(v[:k_star], v[(k_star + 1):]), v[k_star])
            t[v[-1]], q[v[-1]] = self._n_generations, p  # Line 25, 26
        # set for success-based adaptation of mutation strength
        l_w = np.dot(self._w, y_bak[:self.n_parents] > y[:self.n_parents])  # Line 27
        w = self._w_1*w + self._w_2*np.sqrt(self._mu_eff)*(2*l_w - 1)  # Line 28
        self.sigma *= np.exp(norm.cdf(w) - 1.0 + self.a_z)  # Line 29
        return mean, p, w, q, t, v

    def restart_reinitialize(self, args=None, x=None, mean=None, p=None, w=None, q=None,
                             t=None, v=None, y=None, fitness=None):
        is_restart = ES.restart_reinitialize(self)
        if is_restart:
            x, mean, p, w, q, t, v, y = self.initialize(args, is_restart)
            if self.saving_fitness:
                fitness.append(y[0])
            self._print_verbose_info(y)
        return x, mean, p, w, q, t, v, y

    def optimize(self, fitness_function=None, args=None):  # for all generations (iterations)
        fitness = ES.optimize(self, fitness_function)
        x, mean, p, w, q, t, v, y = self.initialize(args)
        if self.saving_fitness:
            fitness.append(y[0])
        self._print_verbose_info(y)
        while True:
            y_bak = np.copy(y)
            # sample and evaluate offspring population
            x, y = self.iterate(x, mean, q, v, y, args)
            if self.saving_fitness:
                fitness.extend(y)
            if self._check_terminations():
                break
            mean, p, w, q, t, v = self._update_distribution(x, mean, p, w, q, t, v, y, y_bak)
            self._n_generations += 1
            self._print_verbose_info(y)
            if self.is_restart:
                x, mean, p, w, q, t, v, y = self.restart_reinitialize(
                    args, x, mean, p, w, q, t, v, y, fitness)
        results = self._collect_results(fitness, mean)
        results['p'] = p
        results['w'] = w
        return results



from unittest import TestCase
import numpy as np

from pypop7.benchmarks import base_functions
from pypop7.benchmarks.utils import generate_xyz, plot_contour


class Test(TestCase):
    def test_generate_xyz(self):
        x, y, z = generate_xyz(base_functions.sphere, [0, 1], [0, 1], num=2)
        self.assertTrue(np.allclose(x, np.array([[0., 1.], [0., 1.]])))
        self.assertTrue(np.allclose(y, np.array([[0., 0.], [1., 1.]])))
        self.assertTrue(np.allclose(z, np.array([[0., 1.], [1., 2.]])))

    def test_plot_contour(self):
        plot_contour(base_functions.sphere, [-10, 10], [-10, 10])
        plot_contour(base_functions.ellipsoid, [-10, 10], [-10, 10])



import numpy as np

from pypop7.optimizers.ga.ga import GA


class G3PCX(GA):
    """G3PCX

    Parameters
    ----------
    problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).

    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`),
              and with the following particular settings (`keys`):
                * 'n_individuals' - population size (`int`, default: `100`),
                * 'n_parents' - parent size (`int`, default: `3`),
                * 'n_offsprings' - offspring size (`int`, default: `2`),
                * 'n_family' - family size (`int`, default: `2`),
                * 'sigma_eta' - distributed variable of PCA (`float`, default: `0.1`),
                * 'sigma_zeta' - distributed variable of PCA (`float`, default: `0.1`),
                * 'mut_eta' - distributed variable of PM (`float`, default: `0.1`),

    Examples
    --------
    Use the GA optimizer `G3PCX` to minimize the well-known test function
    `Rastrigin <http://en.wikipedia.org/wiki/Rastrigin_function>`_:

    .. code-block:: python
       :linenos:

       >>> import numpy
       >>> from pypop7.benchmarks.base_functions import rastrigin  # function to be minimized
       >>> from pypop7.optimizers.ga.g3pcx import G3PCX
       >>> problem = {'fitness_function': rastrigin,  # define problem arguments
       ...            'ndim_problem': 100,
       ...            'lower_boundary': -5 * numpy.ones((100,)),
       ...            'upper_boundary': 5 * numpy.ones((100,))}
       >>> options = {'max_function_evaluations': 1000000,  # set optimizer options
       ...            'n_individuals': 100,
       ...            'n_parents': 3,
       ...            'seed_rng': 2022}
       >>> g3pcx = G3PCX(problem, options)  # initialize the optimizer class
       >>> results = g3pcx.optimize()  # run the optimization process
       >>> # return the number of function evaluations and best-so-far fitness
       >>> print(f"G3PCX: {results['n_function_evaluations']}, {results['best_so_far_y']}")
       G3PCX: 185180, 9.208633855450898e-11

    Attributes
    ----------
    n_individuals : `int`
                    population size.
    n_parents : `int`
                    parent size.
    n_offsprings : `int`
                    offspring size.
    n_family : `int`
                    family size.
    sigma_eta : `float`
                    distributed variable of PCA.
    sigma_zeta : `float`
                    distributed variable of PCA.
    mut_eta : `float`
                    distributed variable of PM.

    References
    ----------
    Kalyanmoy Deb, Ashish Anand, and Dhiraj Joshi.
    A computationally efficient evolutionary algorithm for real-parameter optimization.
    Evolutionary Computation, 10(4):371–395, 2002.
    Part of the code is referenced from:
    https://pymoo.org/algorithms/soo/g3pcx.html
    """
    def __init__(self, problem, options):
        GA.__init__(self, problem, options)
        self.n_offsprings = options.get('n_offsprings', 2)
        self.n_family = options.get('n_family', 2)
        self.sigma_eta = options.get('sigma_eta', 0.1)
        self.sigma_zeta = options.get('sigma_zeta', 0.1)
        self.mut_eta = options.get('mut_eta', 20)
        self.EPSILON = 1e-32
        self.sp_size = self.n_offsprings + self.n_family
        self.indices = np.arange(0, self.n_individuals)
        self.prob_var = min(0.5, 1 / self.ndim_problem)
        self.best_index = None

    def initialize(self, args):
        x = self.rng_initialization.uniform(self.initial_lower_boundary, self.initial_upper_boundary,
                                            size=(self.n_individuals, self.ndim_problem))  # population
        y = np.empty((self.n_individuals,))  # fitness
        for i in range(self.n_individuals):
            y[i] = self._evaluate_fitness(x[i], args)
        self.best_index = np.argmin(y)
        return x, y

    def parent_centric_xover(self, x=None, index=None):
        # calculate the differences from all parents to index parent
        diff_to_index = x - x[index]
        dist_to_index = np.linalg.norm(diff_to_index, axis=-1)
        dist_to_index = np.maximum(self.EPSILON, dist_to_index)

        # find the centroid of the parents
        centroid = np.mean(x, axis=0)

        # calculate the difference between the centroid and the k-th parent
        diff_to_centroid = centroid - x[index]

        dist_to_centroid = np.linalg.norm(diff_to_centroid, axis=-1)
        dist_to_centroid = np.maximum(self.EPSILON, dist_to_centroid)

        # orthogonal directions are computed
        orth_dir = np.zeros_like(dist_to_index)

        for i in range(self.n_parents):
            if i != index:
                temp1 = (diff_to_index[i] * diff_to_centroid).sum(axis=-1)
                temp2 = temp1 / (dist_to_index[i] * dist_to_centroid)
                temp3 = np.maximum(0.0, 1.0 - temp2 ** 2)
                orth_dir[i] = dist_to_index[i] * (temp3 ** 0.5)

        # this is the avg of the perpendicular distances from other parents to the parent k
        D_not = orth_dir.sum(axis=0) / (self.n_parents - 1)
        # generating zero-mean normally distributed variables
        sigma = D_not * np.repeat(self.sigma_eta, self.ndim_problem)
        rnd = np.random.normal(loc=0.0, scale=sigma)

        # implemented just like the c code - generate_new.h file
        inner_prod = np.sum(rnd * diff_to_centroid, axis=-1, keepdims=True)
        noise = rnd - (inner_prod * diff_to_centroid) / dist_to_centroid ** 2

        bias_to_centroid = np.random.normal(0.0, self.sigma_zeta) * diff_to_centroid

        # the array which is finally returned
        return x[index] + noise + bias_to_centroid

    def polynomial_mutation(self, x):
        Xp = x
        mut = np.random.random(self.ndim_problem) < self.prob_var
        mut[self.lower_boundary == self.upper_boundary] = False

        _xl = self.lower_boundary[mut]
        _xu = self.upper_boundary[mut]

        x1 = x[mut]
        eta = np.tile(self.mut_eta, self.ndim_problem)[mut]

        delta1 = (x1 - _xl) / (_xu - _xl)
        delta2 = (_xu - x1) / (_xu - _xl)

        mut_pow = 1.0 / (eta + 1.0)
        rand = np.random.random(x1.shape)
        mask = rand <= 0.5
        mask_not = np.logical_not(mask)

        deltaq = np.zeros(x1.shape)

        xy = 1.0 - delta1
        val = 2.0 * rand + (1.0 - 2.0 * rand) * (np.power(xy, (eta + 1.0)))
        d = np.power(val, mut_pow) - 1.0
        deltaq[mask] = d[mask]

        xy = 1.0 - delta2
        val = 2.0 * (1.0 - rand) + 2.0 * (rand - 0.5) * (np.power(xy, (eta + 1.0)))
        d = 1.0 - (np.power(val, mut_pow))
        deltaq[mask_not] = d[mask_not]

        # mutated values
        _Y = x1 + deltaq * (_xu - _xl)

        # back in bounds if necessary (floating point issues)
        _Y[_Y < _xl] = _xl[_Y < _xl]
        _Y[_Y > _xu] = _xu[_Y > _xu]

        # set the values for output
        Xp[mut] = _Y
        return Xp

    def iterate(self, x=None, y=None, fitness=None):
        indexs = np.random.choice(self.indices, self.n_parents)
        if self.best_index not in indexs:
            indexs[0] = self.best_index

        offsprings = np.empty((self.sp_size, self.ndim_problem))
        off_y = np.zeros((self.sp_size,))
        parent_x = np.empty((self.n_parents, self.ndim_problem))
        for i in range(self.n_parents):
            parent_x[i] = x[indexs[i]]
        parents = np.arange(0, self.n_individuals)

        for i in range(self.n_offsprings):
            offsprings[i] = self.parent_centric_xover(parent_x, i)
            offsprings[i] = np.clip(offsprings[i], self.lower_boundary, self.upper_boundary)
            offsprings[i] = self.polynomial_mutation(offsprings[i])
            offsprings[i] = np.clip(offsprings[i], self.lower_boundary, self.upper_boundary)
            off_y[i] = self._evaluate_fitness(offsprings[i])

        for i in range(self.n_family):
            rindex = np.random.randint(0, self.n_individuals - 1) + i
            if rindex > self.n_individuals - 1:
                rindex = self.n_individuals - 1
            temp = parents[rindex]
            parents[rindex] = parents[i]
            parents[i] = temp
        for i in range(self.n_family):
            offsprings[self.n_offsprings + i] = x[parents[i]]
            off_y[self.n_offsprings + i] = y[parents[i]]
        if self.saving_fitness:
            fitness.extend(off_y)

        order = np.argsort(off_y)
        for i in range(self.n_family):
            x[parents[i]] = offsprings[order[i]]
            y[parents[i]] = off_y[order[i]]
        self.best_index = np.argmin(y)
        return x, y, fitness

    def optimize(self, fitness_function=None, args=None):
        fitness = GA.optimize(self, fitness_function)
        x, y = self.initialize(args=None)
        fitness.extend(y)
        while True:
            x, y, fitness = self.iterate(x, y, fitness)
            if self._check_terminations():
                break
            self._print_verbose_info(y)
            self._n_generations += 1
        return self._collect_results(fitness)



import numpy as np
import matplotlib.pyplot as plt
from celluloid import Camera

from pypop7.benchmarks.utils import generate_xyz
from pypop7.optimizers.core.optimizer import Optimizer
from pypop7.optimizers.rs.rhc import RHC


def plot_contour(func, x, y, levels=None):
    x, y, z = generate_xyz(func, x, y, 500)
    plt.contourf(x, y, z, levels, cmap='bone')
    plt.xlabel('x')
    plt.ylabel('y')


def shi_cd(x):  # fitness function from https://arxiv.org/abs/1610.00040
    return 7 * np.power(x[0], 2) + 6 * x[0] * x[1] + 8 * np.power(x[1], 2)


class RHC1(RHC):
    def optimize(self, fitness_function=None, args=None):
        fitness = Optimizer.optimize(self, fitness_function)
        x_data = []
        best_data = []
        is_initialization = True
        while True:
            if is_initialization:
                x = self.initialize()
                is_initialization = False
            else:
                x = self.iterate()
            y = self._evaluate_fitness(x, args)
            best_data.append(self.best_so_far_x)
            x_data.append(x)
            if self.record_fitness:
                fitness.append(y)
            self._print_verbose_info(y)
            if self._check_terminations():
                break
            self._n_generations += 1
        results = self._collect_results(fitness)
        results['x_data'] = x_data
        results['best_data'] = best_data
        return results


if __name__ == '__main__':
    ndim_problem = 2
    pro = {'fitness_function': shi_cd,
           'ndim_problem': ndim_problem,
           'lower_boundary': -10 * np.ones((ndim_problem,)),
           'upper_boundary': 10 * np.ones((ndim_problem,))}
    opt = {'fitness_threshold': 1e-2,
           'seed_rng': 0,
           'x': np.array([7., -8.]),
           'sigma': 1.0,
           'record_fitness': False,
           'record_fitness_frequency': 1,
           'verbose_frequency': 1}
    solver = RHC1(pro, opt)
    res = solver.optimize()
    print(res)
    fig = plt.figure()
    plt.rcParams['font.family'] = 'Times New Roman'
    plt.rcParams['font.size'] = '12'
    camera = Camera(fig)
    for i in range(len(res['x_data'])):
        plt.xlim([-10, 10])
        plt.ylim([-10, 10])
        plt.xticks([-10, -8, -6, -4, -2, 0, 2, 4, 6, 8, 10])
        plt.yticks([-10, -8, -6, -4, -2, 0, 2, 4, 6, 8, 10])
        plot_contour(shi_cd, [-10, 10], [-10, 10], [0, 10, 100, 500, 1000, 2000])
        plt.scatter(res['best_data'][i][0], res['best_data'][i][1], c='magenta', s=12)
        plt.scatter(res['x_data'][i][0], res['x_data'][i][1], c='limegreen', s=3)
        camera.snap()
    animation = camera.animate()
    animation.save('demo_rhc.gif')



import numpy as np

from pypop7.optimizers.ga.ga import GA


class HRCGA(GA):
    """HRCGA(Hybrid RCGA)

    Parameters
    ----------
    problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).

    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`),
              and with the following particular settings (`keys`):
                * 'n_individuals' - population size (`int`, default: `100`),
                * 'alpha' - step size (`float`, default: `0.7`),
                * 'P_G' - part to run global RCGA (`float`, default: `0.25`),
                * 'N_GM' - best male individuals amount of global RCGA (`int`, default: `400`),
                * 'N_GF' - best female individuals amount of global RCGA (`int`, default: `200`),
                * 'N_LM' - best male individuals amount of local RCGA (`int`, default: `200`),
                * 'N_LF' - best female individuals amount of local RCGA (`int`, default: `200`),

    Examples
    --------
    Use the GA optimizer `HRCGA` to minimize the well-known test function
    `Rastrigin <http://en.wikipedia.org/wiki/Rastrigin_function>`_:

    .. code-block:: python
       :linenos:

       >>> import numpy
       >>> from pypop7.benchmarks.base_functions import rastrigin  # function to be minimized
       >>> from pypop7.optimizers.ga.hrcga import HRCGA
       >>> problem = {'fitness_function': rastrigin,  # define problem arguments
       ...            'ndim_problem': 25,
       ...            'lower_boundary': -5 * numpy.ones((100,)),
       ...            'upper_boundary': 5 * numpy.ones((100,))}
       >>> options = {'max_function_evaluations': 100000,  # set optimizer options
       ...            'n_individuals': 1000,
       ...            'seed_rng': 2022}
       >>> hrcga = HRCGA(problem, options)  # initialize the optimizer class
       >>> results = hrcga.optimize()  # run the optimization process
       >>> # return the number of function evaluations and best-so-far fitness
       >>> print(f"HRCGA: {results['n_function_evaluations']}, {results['best_so_far_y']}")
       HRCGA: 100000, 17.909247914222192

    Attributes
    ----------
    n_individuals : `int`
                    population size.
    alpha: `float`
                    step size.
    P_G: `float`
                    part to run global RCGA.
    N_GM : `int`
                    best male individuals amount of global RCGA.
    N_GF : `int`
                    best female individuals of amount global RCGA.
    N_LM : `int`
                    best male individuals of amount local RCGA.
    N_LF : `int`
                    best female individuals of amount local RCGA.

    References
    ----------
    C. G. Martinez, M. Lozano, F. Herrera, D. Molina, A. M. Sanchez
    Global and local real-coded genetic algorithms based on parent-centric crossover operators
    European Journal of Operational Research 185 (2008) 1088–1113
    """
    def __init__(self, problem, options):
        GA.__init__(self, problem, options)
        self.alpha = options.get('alpha', 0.7)
        self.P_G = options.get('P_G', 0.25)
        self.N_GM = options.get('N_GM', 400)
        self.N_GF = options.get('N_GF', 200)
        self.N_LM = options.get('N_LM', 100)
        self.N_LF = options.get('N_LF', 5)
        self.n_ass = 5
        self.global_function_evaluations = self.P_G * self.max_function_evaluations
        self.global_indices = np.arange(0, self.N_GM)
        self.local_indices = np.arange(0, self.N_LM)

    def initialize(self, args=None):
        x = self.rng_initialization.uniform(self.initial_lower_boundary, self.initial_upper_boundary,
                                            size=(self.n_individuals, self.ndim_problem))  # population
        y = np.empty((self.n_individuals,))  # fitness
        sel_time = np.zeros((self.n_individuals,))
        x_lm = np.empty((self.N_LM, self.ndim_problem))
        x_lf = np.empty((self.N_LF, self.ndim_problem))
        x_gm = np.empty((self.N_GM, self.ndim_problem))
        x_gf = np.empty((self.N_GF, self.ndim_problem))
        g_time = np.empty((self.N_GF,))
        l_time = np.empty((self.N_LF,))
        for i in range(self.n_individuals):
            y[i] = self._evaluate_fitness(x[i], args)
        return x, y, x_lm, x_lf, x_gm, x_gf, sel_time, l_time, g_time

    def ufs(self, x=None, times=None):
        min_time = np.argmin(times)
        return x[min_time], min_time

    def nam(self, x=None, pf=None):
        if self.n_function_evaluations < self.global_function_evaluations:
            places = np.random.choice(self.global_indices, self.n_ass)
        else:
            places = np.random.choice(self.local_indices, self.n_ass)
        similarities = np.empty((self.n_ass,))
        for i in range(self.n_ass):
            similarities[i] = np.linalg.norm(pf - x[places[i]])
        minplace = np.argmax(similarities)
        return x[places[minplace]]

    def iterate(self, x=None, y=None, x_m=None, x_f=None, sel_time=None, times=None, fitness=None, args=None):
        order = np.argsort(y)
        if self.n_function_evaluations < self.global_function_evaluations:
            for i in range(self.N_GM):
                x_m[i] = x[order[i]].copy()
            for i in range(self.N_GF):
                x_f[i] = x[order[i]].copy()
                times[i] = sel_time[order[i]].copy()
        else:
            for i in range(self.N_LM):
                x_m[i] = x[order[i]].copy()
            for i in range(self.N_LF):
                x_f[i] = x[order[i]].copy()
                times[i] = sel_time[order[i]].copy()
        # select parents
        pf, min_time = self.ufs(x_f, times)
        sel_time[order[min_time]] += 1
        pm = self.nam(x_m, pf)

        # create offspring
        I = np.abs(pf - pm)
        l = pf - I * self.alpha
        l = np.clip(l, self.lower_boundary, self.upper_boundary)
        u = pf + I * self.alpha
        u = np.clip(u, self.lower_boundary, self.upper_boundary)
        off_spring = self.rng_initialization.uniform(l, u, size=(self.ndim_problem,))
        off_y = self._evaluate_fitness(off_spring, args)
        if off_y < y[order[-1]]:
            x[order[-1]], y[order[-1]], sel_time[order[-1]] = off_spring, off_y, 0
        if self.saving_fitness:
            fitness.append(off_y)
        return x, y, sel_time, fitness

    def optimize(self, fitness_function=None, args=None):
        fitness = GA.optimize(self, fitness_function)
        x, y, x_lm, x_lf, x_gm, x_gf, sel_time, l_time, g_time = self.initialize(args=None)
        fitness.extend(y)
        while True:
            if self.n_function_evaluations < self.global_function_evaluations:
                x, y, time, fitness = self.iterate(x, y, x_gm, x_gf, sel_time, g_time, fitness, args)
            else:
                x, y, time, fitness = self.iterate(x, y, x_lm, x_lf, sel_time, l_time, fitness, args)
            if self._check_terminations():
                break
            self._print_verbose_info(y)
            self._n_generations += 1
        return self._collect_results(fitness)



"""Repeat Figure 6 in paper
    Latent Action Monto Carlo Tree Search(LA-MCTS)
    Reference
    --------------
    L. Wang, R. Fonseca, Y. Tian
    Learning Search Space Partition for Black-box Optimization using Monte Carlo Tree Search
    NeurIPS 2020
    https://proceedings.neurips.cc/paper/2020/hash/e2ce14e81dba66dbff9cbc35ecfdb704-Abstract.html
    Code of paper:
    https://github.com/facebookresearch/LaMCTS/tree/main/LA-MCTS
"""
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
import pickle

from pypop7.benchmarks.base_functions import ackley, rosenbrock
from pypop7.optimizers.bo.lamcts import LAMCTS as Solver

sns.set_theme(style='darkgrid')


def read_pickle(function, ndim):
    file = function + '_' + str(ndim) + '.pickle'
    with open(file, 'rb') as handle:
        result = pickle.load(handle)
        return result


def write_pickle(function, ndim, result):
    file = open(function + '_' + str(ndim) + '.pickle', 'wb')
    pickle.dump(result, file)
    file.close()


def plot(function, problem_dim):
    result = read_pickle(function, problem_dim)
    print("1")
    plt.figure()
    if function == "ackley":
        plt.ylim([0, 20])
        plt.yticks([0, 5, 10, 15])
    elif function == "rosenbrock":
        plt.yscale('log')
        plt.ylim([1e1, 1e8])
        plt.yticks([1e2, 1e3, 1e4, 1e5, 1e6, 1e7])
    if problem_dim == 20:
        plt.xlim([0, 1000])
        plt.xticks([0, 200, 400, 600, 800, 1000])
    elif problem_dim == 100:
        plt.xlim([0, 10000])
        plt.yticks([0, 2e3, 4e3, 6e3, 8e3, 1e4])
    plt.plot(result['fitness'][:, 0], result['fitness'][:, 1], color='orange')
    plt.xlabel("# samples")
    plt.ylabel("f(x)")
    plt.title(function.capitalize() + "-" + str(problem_dim) + "d")
    plt.show()


if __name__ == "__main__":
    for f in [ackley, rosenbrock]:
        for d in [20, 100]:
            print('*' * 7 + ' ' + f.__name__ + ' ' + '*' * 7)
            problem = {'fitness_function': f,
                       'ndim_problem': d,
                       'lower_boundary': -5 * np.ones((d,)),
                       'upper_boundary': 10 * np.ones((d,))}
            options = {'max_function_evaluations': 1000,
                       'fitness_threshold': 1e-5,
                       'seed_rng': 0,
                       'n_individuals': 40,  # the number of random samples used in initialization
                       'Cp': 1,  # Cp for MCTS
                       'leaf_size': 10,  # tree leaf size
                       'kernel_type': "rbf",  # used in SVM
                       'gamma_type': "auto",  # used in SVM
                       'solver_type': 'bo',  # solver can be bo or turbo
                       'verbose': 1,
                       'saving_fitness': 1}
            if f.__name__ == "rosenbrock":
                problem['lower_boundary'] = -10 * np.ones((d,))
            if d == 100:
                options['max_function_evaluations'] = 1e4
            solver = Solver(problem, options)
            results = solver.optimize()
            print(results)
            write_pickle(f.__name__, d, results)
            plot(f.__name__, d)



"""This code only tests whether the corresponding optimizer could run or not.
  In other words, it *cannot* validate the correctness regarding its working procedure.
  For the correctness validation of programming, refer to the following link:
    https://github.com/Evolutionary-Intelligence/pypop/wiki/Correctness-Validation-of-Optimizers.md
"""
import unittest
import time
import numpy as np

from pypop7.benchmarks.base_functions import ackley
from pypop7.optimizers.bo.lamcts import LAMCTS as Solver


class TestLAMCTS(unittest.TestCase):
    def test_optimize(self):
        start_run = time.time()
        ndim_problem = 20
        for f in [ackley]:
            print('*' * 7 + ' ' + f.__name__ + ' ' + '*' * 7)
            problem = {'fitness_function': f,
                       'ndim_problem': ndim_problem,
                       'lower_boundary': -5 * np.ones((ndim_problem,)),
                       'upper_boundary': 10 * np.ones((ndim_problem,))}
            options = {'max_function_evaluations': 1e2,
                       'fitness_threshold': 1e-5,
                       'seed_rng': 0,
                       'n_individuals': 40,  # the number of random samples used in initialization
                       'Cp': 1,  # Cp for MCTS
                       'leaf_size': 10,  # tree leaf size
                       'kernel_type': "rbf",  # used in SVM
                       'gamma_type': "auto",  # used in SVM
                       'solver_type': 'bo',  # solver can be bo or turbo
                       'verbose': 1,
                       'saving_fitness': 1}
            solver = Solver(problem, options)
            results = solver.optimize()
            print(results)
            print('*** Runtime: {:7.5e}'.format(time.time() - start_run))


if __name__ == '__main__':
    unittest.main()
